package provide hooks 2.0

set Hooks {}
set HookCounter 0

# Hook management commands
proc hook {events Handler {filter 0}} {
	set ns [uplevel 1 namespace current]
	foreach ev $events { 
		lappend ::Hooks $ev $Handler $ns $filter
	}
}

proc unhook {events Hook {space {}}} {
	if {[string equal $space {}]} {
		set space [uplevel 1 namespace current]
	}
	set newHooks [list]
	foreach Name $events {
		foreach {name hook context filter} $::Hooks {
			if {![string match $Name $name] ||
			    ![string match $Hook $hook] ||
			    ![string match $space $context]} {
				lappend newHooks $name $hook $context $filter
			}
			#else {puts "Removed: $name ${context}::$hook"}
		}
	}
	set ::Hooks $newHooks
}

proc Event {name args} {
	global Hooks
	set lst [list]
	foreach {ev ha ns fi} $Hooks {
		if {[string match $ev $name]} {lappend lst [list $ha $ns $fi]}
	}
	foreach item [lsort -command sortEvents $lst] {
		foreach {handler ns filter} $item break
		if {[catch {set res [namespace eval $ns $handler $args]} v]&&
		    ![string equal $name Error:Module]} {
			Event Error:Module ${ns}::$handler $v
		} elseif {$filter} {
			if {[llength $res]==[llength $args]} { 
				set args $res
			} else {
				Event Error:Module ${ns}::${handler} {
					Number of return values of filter
					does not match number of arguments
				}
			}	
		}
	}
	set args
}

proc sortEvents {id1 id2} {
	string compare [lindex $id1 0] [lindex $id2 0]
}

proc PEvent {Name id oldval newval} {
	foreach {Hook Handler ns fi} $::Hooks {
		if {[string match $Hook $Name]} { 
			if {[catch {${ns}::$Handler $id $oldval $newval} v] } {
				Event Error:Module ${ns}::$Handler $v
			}
		}
	}
}

foreach {name flag} {handler 0 filter 1} {
	set body {
		if {[string equal $id {}]} {
			set id 50hook[clock seconds][incr ::HookCounter]
		}
		set ns [uplevel 1 namespace current]
		proc ${ns}::${id} $parms $body
		namespace eval $ns [list hook $events $id }
	append body "$flag\]"
	proc $name {events parms body {id {}}} $body
}

set flashcnt 0
proc Flash {type} {
	switch -- $type {
		more {if {[incr ::flashcnt]==1} {Flasher}}
		less {if {![incr ::flashcnt -1]} {
				after cancel $::flasher
				Event Flashing 1
			}	
		}
		* {return -code error "Unknown parameter $type for Flash"}
	}
}

proc Flasher {{stage 0}} {
	Event Flashing $stage
	set ::flasher [after 300 "Flasher [expr $stage^1]"]
}

