package provide AlicqMaster 0.8

set licqconfig ignore
set modlist {encoding antispam history saver geometry base menu whitepages colorer url tkonfig}

proc defaults {} {
	set ::version 0.8
	set ::fincmd {destroy .}
	set ::cancmd {destroy .}
	switch $::tcl_platform(platform) {
		unix {
			set ::BASE ~/.alicq
			set ::RCLIST [list [file join $::BASE alicqrc]]
		}
		windows {
			set ::BASE [file join $::env(USERPROFILE) alicq]
			set ::RCFILE [list [file join $::BASE alicq.cfg]]
		}
	}
}

proc FrameFirst {content} {
	set f [frame $content.first]		
	label $content.first.lb -text "This master will help you to ctreate\n\
				basic configuration."\
			-height 10 -width 50
	pack $content.first.lb
	return [list $f "Welcome to Alicq $::version"]
}


proc FrameLicq {content} {
	set f [frame $content.licq]
	set ::licqconfig convert
	radiobutton $f.use -text "Use Licq configuration files (deprecated)" -value use -variable licqconfig -anchor w -state disabled
	radiobutton $f.convert -text "Convert them to Alicq config files" -value convert -variable licqconfig -anchor w
	radiobutton $f.ignore -text "Don't use Licq configuration" -value ignore -variable licqconfig -anchor w
	pack $f.use $f.convert $f.ignore -fill x
	return [list $f "Alicq found configuration files of Licq.\nDo you want to use them?"]
				
}

proc FrameUser {content} {
	set f [frame $content.user]
	set inf [frame $f.info -borderwidth 2 ]
	set ::HaveUIN 1
	set :uin 234
	grid [label $inf.lbUIN -text "ICQ UIN"] [set enUIN [entry $inf.enUIN -textvariable ::uin]]
	grid [label $inf.lbPwd1 -text "Password"] [entry $inf.enPwd1 -show * -textvariable ::pwd1]
	grid [label $inf.lbPwd2 -text "Re-type wassword"] [entry $inf.enPwd2 -show * -textvariable ::pwd2]
	radiobutton $f.existing -text "I already have registered ICQ UIN" -value 1 -variable ::HaveUIN -command "$enUIN configure -state normal" -anchor w 
	radiobutton $f.new -text "Register me new UIN" -value 0 -variable ::HaveUIN -command "$enUIN configure -state disabled" -anchor w -state disabled
	pack $f.existing $f.new -fill x
	pack $inf -pady 5 -padx 5
	return [list $f "Register ICQ UIN"]
}

proc FrameLast {content} {
	set f [frame $content.last]
	label $f.lb -height 5 -text "Configuration files were updated.\nClick finish."
	pack $f.lb
	return [list $f "Configuration master updated you config file"]
}
proc UpdateFiles {content} {
	set f [frame $content.update]
	label $f.lb -height 5 -text "Now configuration files will be updated."
	pack $f.lb
	return [list $f "Ready to update configuration files"]
}


proc DisplayContent {label content btn idx} {
	set frame [lindex $::Frames $idx] 
	$btn.back configure -state [expr ($idx)?"normal":"disabled"]\
		-command "DisplayContent $label $content $btn [expr $idx-1]"
	$btn.next configure \
		-command "if {!\[Validate $frame\]} {DisplayContent $label $content $btn [expr $idx+1]}"
	if {$idx==[expr [llength $::Frames]-1]} {
		foreach b {back cancel} {$btn.$b configure -state disabled}
		$btn.next configure -text "Finish" -command $::fincmd
	}
	foreach slave [pack slaves $content] {pack forget $slave}
	pack $frame -expand yes
	$label.lb configure -text [lindex $::Labels $idx]
}

proc Validate {frame} {
	if {$frame==".content.licq"} {
		if {$::licqconfig=="convert"} {
			package require convert
			puts $LicqConvert::User
			array set uinfo $LicqConvert::User
			set ::uin $uinfo(Global:ICQ:Info|Uin)
			set ::pwd1 $uinfo(Global:ICQ:Info|Password)
			set ::pwd2 $::pwd1
		}
	}
	if {$frame==".content.user"} {
		#set ::uin [$frame.info.enUIN get]
		if {![regexp {^\d+$} $::uin]} {
			set ::error "UIN should contain only digits"
			return 1
		}
		#set ::pwd [$frame.info.enPwd1 get]
		#set pwd2 [$frame.info.enPwd2 get]
		if {$::pwd1!=$::pwd2} {
			set ::error "Passwords do not match"
			return 1
		}
	}
	if {$frame==".content.update"} {
		return [WriteConfig]
	}
	set ::error {}
	return 0
}

proc WriteConfig {} {
	set base $::BASE
	set rcfile [lindex $::RCLIST 0]
	catch { 
		if {![file exists $base]} {
			file mkdir $base
			file attributes 0700
		}
		set fd [open $rcfile w 0600]
		puts $fd "# Autogenerated by master\n"
		puts $fd "Module icqconn.tcl"
		puts $fd "pset Global:ICQ:Info|Uin $::uin"
		puts $fd "pset Global:ICQ:Info|Password $::pwd1"

		puts $fd "\n#Modules"
		foreach item $::modlist {
			puts $fd "Module ${item}.tcl"
		}
		puts $fd "\n# Source contacts file"
		set contacts [file join $base contacts]
		puts $fd "set contacts {$contacts}"
		puts $fd "source \$contacts"
		puts $fd "pset Global:Saver|paths \[list Contact:* \$contacts Group:* \$contacts * \$::RCFILE\]"
		close $fd
		# Convert licq config if needed 
		if {$::licqconfig=="convert"} {
			set fd [open $contacts w 0600]
			foreach line $LicqConvert::Contacts { puts $fd $line }
			close $fd
			if {![file exists [file join $base history]]} {
				file copy ~/.licq/history [file join $base]
			}	
		}
		if {![file exists $contacts]} {
			set fd [open $contacts w 0600]
			puts $fd "# Put contacts here"
			close $fd
		}
	} ::error
}

proc MainDialog {top} {
	wm title . {Alicq configuration master}
	option add *Frame.relief groove
	option add *Entry.relief groove
	option add *Entry.Background gray95
	pack [frame $top.label ] -fill x
	pack [label $top.label.lb -height 3 -bg white -fg blue] -fill both -expand yes
	pack [frame $top.content -borderwidth 1 -width 400 -height 200] -fill both -expand yes
	pack [label $top.error -height 1 -bg white -fg red -textvariable ::error] -fill both -expand yes
	pack [frame $top.btn] 
	button $top.btn.next -text "Next" 
	button $top.btn.cancel -text "Cancel" -command $::cancmd
	button $top.btn.back -text "Back" -state disabled
	pack $top.btn.back $top.btn.cancel $top.btn.next -padx 10 -side left
	set idx 0
	set lst {FrameFirst FrameLicq FrameUser UpdateFiles FrameLast}
	if {![file exists ~/.licq]} {
		set pos [lsearch $lst FrameLicq]
		set lst [lreplace $lst $pos $pos]
	}
	foreach frame $lst {
		foreach {id txt} [$frame $top.content] {
			lappend ::Frames $id
			lappend ::Labels $txt
		}
		incr idx
	}
	pack propagate $top.content false
	DisplayContent $top.label $top.content $top.btn 0
}
if {[info exists ::version]} {
	set ::fincmd {foreach slave [pack slaves .] {destroy $slave}; set z _}
	set ::cancmd {destroy .}
	bind . <Destroy> {set z _}
	MainDialog {}
	vwait ::z
} else { 
	defaults
	MainDialog {}
}

