# This module is part of Alicq instant messenger. 
# It is used to encrypt outgoing and decrypt incoming messages from contacts.
# 
# Author: Ihar Viraheichyk

# PGP encryption. Strong, but wery slow and works only with
# external gpg program.
namespace eval PGP {

property::Add Global:Crypt:PGP|Passphrase {
	-type string -save change
	-descr {Passphrase to unlock secret PGP key}
	}
if {![info exists ::env(TMPDIR)]} { set ::env(TMPDIR) /tmp }

proc Encode {uid message} {
	set tmp [file join $::env(TMPDIR) alicq.[pid]]
	if {[catch {
		set f [open "|gpg --batch --yes -r [pget $uid|crypt:key] --armour -o $tmp -e" w]
		puts $f $message; close $f
		set f [open $tmp r]; set message [read -nonewline $f]; close $f
		#catch {file delete $tmp}
	} v]} {Event Error 0:80 "Encryption failed: $v"}
	return $message
}

proc Decode {uid message} {
	set tmp [file join $::env(TMPDIR) alicq1.[pid]]
	set tmp2 [file join $::env(TMPDIR) alicq2.[pid]]
	set f [open $tmp w]; puts $f $message; close $f
	if [file exists $tmp2] {file delete $tmp2}
	set output "Error decoding message"
	catch {
		set f [open "|gpg --batch --no-tty --yes --passphrase-fd 0 -r [pget $uid|crypt:key] --armour -o $tmp2 -d $tmp" w]
		puts $f [pget Global:Crypt:PGP|Passphrase]; close $f
		set f2 [open $tmp2 r]; set output [read -nonewline $f2]; close $f2
		puts "OUT: $output"
	}
	#catch {file delete $tmp $tmp2}
	return $output
}

proc Detect {message} {
	regexp -- {-----BEGIN PGP MESSAGE-----.*-----END PGP MESSAGE-----} $message
}

}
# Bitwise XOR encryption. Very simple, fast and weak, but sufficient 
# for most cases
namespace eval XOR {

# This property contains offset in key for XOR method
#property::Add	Contact:ICQ:*|crypt:offset {-type integer -default 0}

# XOR two strings 
proc Xor {key start message} {
	if {![string length $key]} {return $message}
	set result {}
	for {set pos 0} {[binary scan $message @${pos}c c1]} {incr pos} {
		if {![binary scan $key @${start}c c2]} {
			set start 0; binary scan $key @${start}c c2
		} else {incr start}
		append result [binary format c [expr $c1^$c2]]
	}
	set result
}

proc Decode {uid message} {
	regexp "^---XOR begin:(\[0-9\]*)---\n(.*)\n---XOR end---$" $message s offset message
	encoding convertfrom utf-8 [Xor [pget $uid|crypt:key] $offset\
					[binary format H* $message]]
}
proc Encode {uid message} {
	set offset 0
	set result [Xor [pget $uid|crypt:key] $offset\
			[encoding convertto utf-8 $message]]
	binary scan $result H* result 
	return 	"---XOR begin:$offset---\n$result\n---XOR end---"
}
proc Detect {message} {regexp -- {^---XOR begin:[0-9]*---.*---XOR end---$} $message}

}

# Detect encryption method and decode message
filter Incoming {type uid time message} {
	if {$type!={text}} return
	foreach method [namespace children] {
		if {[llength [info procs ${method}::Detect]] &&\
		    [${method}::Detect $message]} {
			return [list $type $uid $time\
			        [${method}::Decode $uid $message]]
		}
	}
	list $type $uid $time $message
} 10DecryptIncoming

# Encrypt message if encryption method is set
filter Outgoing {type uid message} {
	if {[pget $uid|crypt:method]!={none}&&$type=={text}} {
		set message [[pget $uid|crypt:method]::Encode $uid $message]
	}
	list $type $uid $message
} 90EncryptOutgoing

property::Add	Contact:*|crypt:key		{
	-type string -weight 70:02 -default {no} -name {Cryptography:Key}
	-save change -hook Crypt:KeyChanged
	}
property::Add	Contact:*|crypt:method	[list\
	-type variant -weight 70:01 -default none -name {Cryptography:Method}\
	-values [concat none [map x [namespace children] {namespace tail $x}]]\
	-save change -hook Crypt:MethodChanged]

