# This module provides basic dialogs of Alicq
#
# Author: Igor Vergeichik

package require BWidget

option add *timeFormat {%X %x} widgetDefault

proc UserDialog {uid time {text {}}} {
	set top ".msg-[string map {: -} $uid]"
	# Multicast groups
	set wins [list]
	if {$text!={}} {set wins [DetectWindows $uid .msg]}
	if {![llength $wins] && ![winfo exists $top]} {
		toplevel $top -class AlicqUserWindow
		wm title $top "[pget $uid|Alias] \([lrange [split $uid :] 1 end]\)"
		grid [set pw [PanedWindow $top.pw -side right]] -sticky news
		set fr [$pw add -weight 1]
		bindtags $fr [concat MsgFrame [bindtags $fr]]
		set in [ScrolledTxt $fr.in -state disabled -wrap word]
		set fr [$pw add -weight 1]
		bindtags $fr [concat MsgFrame [bindtags $fr]]
		set out [ScrolledTxt $fr.out -wrap word]
		bindtags $in.txt [concat MessageText [bindtags $in.txt]]
		pack $in $out -fill both
		bind $out.txt <<Send>> "[nc SendMessage $uid $pw];break"
		grid [ButtonBar $top.btn [list\
			[list Send [nc SendMessage $uid $pw]]\
			[list Cite [nc CiteMessage $uid $out.txt] <<Cite>>]\
			[list History [list Event $uid|history $uid]]\
			[list Info "Event $uid|info:view $uid"]\
			[list Close "destroy $top" <<Close>>]\
		]] -sticky s 
		grid rowconfigure $top 0 -weight 1
		grid columnconfigure $top 0 -weight 1
		set wins [list $top]
	}
	if {$text!={}} {
		foreach top $wins {
			DisplayMessage incoming [$top.pw getframe 0].in.txt\
				$uid $text $time
		}
		pset $uid|LastMessage $text
	}
	raise $top
	focus [$top.pw getframe 1].out.txt
}

proc CiteMessage {uid out} {
	set msg [property::Get $uid|LastMessage]
	foreach ln [split $msg "\n"] {
		$out insert end "> $ln\n"
	}
}
proc SendMessage {uid pane} {
	set out [$pane getframe 1].out.txt
	set in [$pane getframe 0].in.txt
	set message [$out get 1.0 "end -1 chars"]
	if {$message=={}} return
	if {[string match Contact:* $uid]} {
		Event Outgoing text $uid $message
	} elseif {[string match Group:common:* $uid]} {
		foreach y [grep x [lsprop Contact:*|Groups] {
			[lsearch -exact [pget $x] $uid]!=-1
			}] {
			Event Outgoing text [lindex [split $y |] 0] $message
		}
	}
	DisplayMessage outgoing $in Global:ICQ:Info $message [clock seconds]
	$out delete 1.0 end
	focus $out
}

proc DisplayMessage {type txt uid message time} {
	$txt configure -state normal
	set fmt [option get $txt timeFormat TimeFormat]
	set header "\[[clock format $time -format $fmt]\] [pget $uid|Alias]:\n"
	set idx1 [$txt index "end -1 char"]
	$txt insert end $header
	$txt tag add header $idx1 [$txt index "end -1 char"]
	$txt insert end "$message\n\n"
	$txt tag add $type $idx1 [$txt index "end -1 char"]
	$txt see end
	$txt configure -state disabled
}

# New objects dialogs
hook Global:Actions|add:group {NewObjDialog group ::AddObject {Enter group name:}}
hook Global:Actions|add:contact {NewObjDialog contact ::AddObject {Enter UIN:}}

proc NewObjDialog {type action caption args} {
	set top ".new-$type"
	if {[catch {toplevel $top}]} return
	wm title $top "New $type"
	label $top.lb -text $caption
	entry $top.en
	pack $top.lb $top.en -fill x
	append action " $type \[$top.en get\]; destroy $top"
	pack [ButtonBar $top.buttons [list\
		[list Create $action <<Accept>>]\
		[list Close "destroy $top" <<Close>>]]] -side bottom
}

# Operations with contacts
handler IncomingContacts {uid ContactList} {
	set UIN [lindex [split $uid :] 2]
	set top ".contacts-$UIN"
	if {[catch {raise $top}]} {
		toplevel $top -class AlicqContactsWindow
		wm title $top "Contacts from [pget $uid|Alias] ($UIN)"

		grid [ListBox $top.list -dragenabled 1] -sticky news
		$top.list bindText <1> [list $top.list selection add]
		$top.list bindText <3> [list $top.list selection remove]
		grid [ButtonBar $top.btn [list\
			[list Add [nc AddIncomingContacts $top.list]]\
			[list Close "destroy $top" <<Close>>]\
		]]
		grid rowconfigure $top 0 -weight 1
		grid columnconfigure $top 0 -weight 1
	}
	foreach {uin alias} $ContactList {
		$top.list insert end $uin\
			-text "$alias (ICQ $uin)"\
			-image img:offline -data $alias
	}
}
proc AddIncomingContacts {list} {
	foreach x [$list selection get] {
		if {![llength [lsprop Contact:ICQ:$x|*]]} {
			Contact $x [$list itemcget $x -data] {other}
		} 
		$list delete $x
	}
	if {![llength [$list items]]} { destroy [winfo toplevel $list]}
}

# Authorization handling
handler AuthorizationRequest {uin time data} {
	set top ".auth-$uin"
	if {![catch {raise $top}]} return
	toplevel $top -class AlicqAuthRequestWindow
	wm title $top "Authorization request from $uin"
	pack [set f [frame $top.f -borderwidth 1]] -expand yes -fill both
	foreach key {uin nick fname lname email xxx reason} value [concat $uin $data] {
			set lb [label $f.label$key -text $key]
			set en [entry $f.entry$key]
			$en insert 0 $value
			$en configure -state disabled
			grid $lb $en
	}
	pack [ButtonBar $top.btn [list\
		[list Authorize "icq::Authorize $uin 1; ::AddObject contact $uin; destroy $top" <<Accept>>]\
		[list Reject "icq::Authorize $uin 0; destroy $top" {}]\
		[list Cancel "destroy $top" <<Close>>]\
	]]
}

proc AskAuthorization {uid} { 
	set top ".askauth-[string map {: -} $uid]"
	set uin [lindex [split $uid :] 2]
	if {[catch {toplevel $top -class AlicqAskAuthWindow}]} return
	wm title $top "Beg authorization from $uid"
	pack [label $top.lb -text "Authorization request"] -fill x
	pack [set in [ScrolledTxt $top.in -width 30 -height 4]]\
		-expand yes -fill both
	pack [ButtonBar $top.btn [list\
		[list Send "icq::AskAuthorization $uin \[$in.txt get 1.0 end\]; destroy $top" <<Send>>]\
		[list Cancel "destroy $top" <<Close>>]\
	]]
}


# Password change dialog
proc ChangePasswordDlg {} {
	set top .password
	if [catch {toplevel $top}] return
	wm title $top "Changing ICQ password"
	pack [frame $top.f -borderwidth 1 -relief raised] -fill both -pady 2
	foreach {i msg} {1 Enter 2 Re-type} {
		label $top.f.lb${i} -text "$msg new password"
		entry $top.f.en${i} -show *
		grid $top.f.lb${i} $top.f.en${i} 
	}
	pack [ButtonBar $top.btn [list\
		[list Change [nc ChangePassword $top] <<Accept>>]\
		[list Cancel "destroy $top" <<Close>>]]]
}

proc ChangePassword {top} {
	set pwd1 [$top.f.en1 get]
	set pwd2 [$top.f.en2 get]
	if {$pwd1!=$pwd2} { 
		MessageDlg .z -icon error -type ok\
			-message "Passwords do not match. Reenter"
		return	
	}
	set ref [expr [clock seconds]& 0xFF]
	handler PasswordChanged {ref} "
		if {$ref!=\$ref} return
		MessageDlg .z -icon info -type ok\
			-message {Your ICQ password was changed.\nUpdate your cnofig file}
		unhook PasswordChanged pc-$ref
	" pc-$ref
	icq::ChangePassword $pwd1 $ref
	destroy $top
}

# URL dialog
proc SendURL {uid} {
	set top ".url-[string map {: -} $uid]"
	set uin [lindex [split $uid :] 2]
	if {[catch {toplevel $top -class AlicqURLWindow}]} return
	wm title $top "Sending URL to $uid"
	pack [label $top.lb1 -text "Description"] -fill x
	pack [set in [ScrolledTxt $top.in -height 4]] -fill both -expand yes
	pack [label $top.lb2 -text "URL"] -fill x
	pack [entry $top.url] -fill x
	pack [ButtonBar $top.btn [list\
		[list Send "icq::message URL $uin \[list \[$in.txt get 1.0 end\] \[$top.url get\]\]; destroy $top" <<Send>>]\
		[list Cancel "destroy $top" <<Close>>]\
	]]
}

# SMS dialog
proc SendSMS {uid} {
	set top ".sms-[string map {: -} $uid]"
	if {[catch {toplevel $top -class AlicqSMSWindow}]} return
	wm title $top "Sending SMS"
	pack [label $top.lb1 -text "Message"] -fill x
	pack [set in [ScrolledTxt $top.in -height 4]] -fill both -expand yes
	pack [label $top.lb2 -text "Modile number"] -fill x
	pack [entry $top.sms] -fill x
	pack [ButtonBar $top.btn [list\
		[list Send "icq::SendSMS \[$top.sms get\] \[$in.txt get 1.0 end\];destroy $top" <<Send>>]\
		[list Cancel "destroy $top" <<Close>>]\
		]]
}

proc SendContacts {uid} {
	set top ".contacts-$uid"
	if {![catch {raise $top}]} return
	toplevel $top -class AlicqContactsWindow
	wm title $top "Contacts for $uid"
	set lbox [ListBox $top.list -deltay 20 -dropenabled yes\
		-droptypes {TREE_NODE} -dropcmd [nc Drop]]
	$lbox insert end zz -fill blue -text "Drad contacts to send here"
	grid $lbox -sticky news
	grid [ButtonBar $top.btn [list\
		[list Send "icq::message Contacts [lindex [split $uid :] 2] \[[nc SendContactlist $lbox]\];destroy $top" <<Send>>]\
		[list Cancel "destroy $top"]\
	]]
	grid rowconfigure $top 0 -weight 1
	grid columnconfigure $top 0 -weight 1
}

proc SendContactlist {name} {
	set res [list]
	foreach uid [$name items 1 end] {
		lappend res [lindex [split $uid :] 2] [pget $uid|Alias]
	}
	set res
}

proc Drop {widget src dest op datatype data} {
	set uid [tree::UidByNode $data]
	if {[string match Group:* $uid]} return
	if {[$widget exists $uid]} return
	$widget insert end $uid\
		-text [pget $uid|Node:Text]\
		-image img:[pget $uid|Node:Image]
}

handler Contact:*|history {id} {
		set uid [lindex [split $id |] 0]
		set uin [lindex [split $uid :] 2]
		set top ".history-[string map {: -} $uid]"
		if {[catch {toplevel $top -class AlicqHistoryWindow}]} return
		set in [ScrolledWindow $top.in]
		text $in.txt 
		bindtags $in.txt [concat MessageText [bindtags $in.txt]]
		$in setwidget $in.txt
		button $top.ok -text "Ok" -command "destroy $top"
		bind $top <<Close>> "destroy $top"
		pack $top.in -expand yes -fill both
		pack $top.ok
		
		set cmd "FillHistory $in.txt $uid"
		hook HistoryItem $cmd
		bind $top <Destroy> [namespace code "
			if {{$top}=={%W}} { unhook HistoryItem {$cmd} }
		"]
		Event HistoryRequest $uin
}

proc FillHistory {txt uid uin type time msg} {
	if {![string equal $uid Contact:ICQ:$uin]} return
	if {$type=="outgoing"} { set uid Global:ICQ:Info }	
	set idx1 [$txt index "end -1 char"]
	DisplayMessage $type $txt $uid $msg $time
	set idx2 [$txt index "end -1 char"]
	#Event DisplayFilter $txt $idx1 $idx2
}

handler {*|Send:*} {id {time {}} {msg {}}} {
	foreach {uid action} [split $id |] break
	set pending [list]
	if {[string match Contact:* $uid]} { set pending [pget $uid|Pending] }
	if {[llength $pending]} {
		foreach {type time message} $pending {
			SimpleWrapper Send:Message $type $uid $time $message
		}
		pset $uid|Pending [list]
	}
	SimpleWrapper $action text $uid $time $msg
}

proc SimpleWrapper {action type uid time msg} {
	switch -- $type {
		{descr URL} {
			set msg "URL: [lindex $msg 1]\nDescription: [lindex $msg 0]"
		}
	}
	switch -glob -- $action {
		Send:Message	{UserDialog $uid $time $msg}
		Send:SMS	{SendSMS $uid}
		Send:URL	{SendURL $uid}
		Send:Auth	{icq::Authorize [lindex [split $uid :] 2] 1}
		Send:AuthReq	{AskAuthorization $uid}
		Send:Contacts	{SendContacts $uid}
	}
}

# If alias changed, change message window caption as well
handler *|Alias {id oldval newval} {
	set uid [lindex [split $id |] 0]
	if {[winfo exists .msg:$uid]} {
		wm title .msg:$uid "$uid \($newval\)"
	}
}

proc DetectWindows {uid {prf .msg}} {
	if {[winfo exists [set w $prf-[string map {: -} $uid]]]}\
		{return [list $w]}
	grep y [map x [pget $uid|Groups] { set _ $prf-[string map {: -} $x]}]\
		{[winfo exists $y]}
}

# block messages
handler Incoming {type uid time message} {
	if {[string is true [pget Global:UI:Messages|Popup]] ||\
	    [llength [DetectWindows $uid]]} {
		Event $uid|Send:Message $uid|Send:Message $time $message
	} else {	
		lappend [property::Variable $uid|Pending]\
			$type $time $message
	}
}

bind MsgFrame <Destroy> {+
	option add *[string range %W 1 end].height [winfo height %W]
}

option add *Frame.height 100 widgetDefault

property Global:Actions|add:contact {-type action -name {Add:Contact} -hook Add:Contact}
property Global:Actions|add:group	{-type action -name {Add:Group} -hook Add:Group}
property Contact:*|history		{-type action -name History -hook ShowHistory}

property Contact:*|Pending	{-type list -default {} -hook PendingChanged}

property Global:UI:Messages|Popup	{-type boolean -default 1 -name {Interface:Popup on incoming messages} -scope {mainmenu} -save change -hook PopupChanged}
property Contact:*|LastMessage	{-type string -default {}}
		
property Group:common:*|Send:Message	{-type action -weight 01:01 -name {Multicast:Message} -hook Send:Message}

