# This Alicq module used to save positions of newly opened wnidows and open
# then with saved geometry on creation

# Author: Ihar Viarheichyk <iverg@mail.ru>

variable file [file join $::BASE geometry]

rename ::toplevel ::toplevel.orig

proc ::toplevel {pathname args} {
	set top [uplevel toplevel.orig $pathname $args]
	if {[set gm [option get $pathname geometry Geometry]]!={}} {
		wm geometry $pathname $gm
	}
	set top
}

proc Save {window} {
	variable file
	set gm [winfo geometry $window]
	if {$window=={.}} {
		set key "[winfo class .]\\\.geometry:"
		set value "[winfo class .].geometry: $gm"
	} else {	
		regsub {\.} $window {} window
		set key "\\\*${window}\\\.geometry:"
		set value "*${window}.geometry: $gm"
		option add *${window}.geometry $gm
	}	
	property::Save $file $key $value
}

proc bindclasses {list} {
	foreach class $list {
		bind $class <Destroy> [nc Save %W]
	}
}

bind MsgFrame <Destroy> +[namespace code {
	set height [winfo height %W]
	set win [string range %W 1 end]
	set value "*${win}.height: $height"
	set key "\\\*${win}\\\.height:"
	property::Save $file $key $value
}]


# Add default geometry for some classes
foreach {class geom} {AlicqUserWindow "500x300" 
		      AlicqHistoryWindow "600x400"} {
	option add "*${class}.geometry" $geom widgetDefault
}
if [file exists $file] {
	if {[catch { option readfile $file widgetDefault }]} {
		.file rename -force $file ${file}.bad
	}
}

bindclasses {AlicqUserWindow AlicqContactsWindow AlicqAuthRequestWindow \
			 AlicqAskAuthWindow AlicqURLWindow AlicqSMSWindow\
			 AlicqHistoryWindow}
			 
after idle [namespace code {bindclasses [winfo class .]}]
