# Binding for ICQ protocol library

variable proxy {}
variable contacts [list]

# Contact list management commands 
proc ContactInfo {uid myref ref info} {
	if {$ref!=$myref} return
	foreach {key val} $info { if {$key=="Nick"} { pset $uid|Alias $val } }
	unhook ContactInfo "ContactInfo $uid $myref"
}

handler AddICQContact {uin groups alias} {
	set uin [string trim $uin]
	Contact $uin $alias $groups
	icq::AddToList contacts [list $uin]
	set ref [expr [clock seconds] & 0xff]
	hook ContactInfo "ContactInfo Contact:ICQ:$uin $ref"
	icq::ShortInfoRequest $uin $ref
}

handler IncomingMessage {uin time message} {
	if {![llength [lsprop Contact:ICQ:$uin|Alias]]} {
		Event AddICQContact $uin {other} $uin
	}	
}

handler Error:Protocol {handler reason dump args} {
	binary scan $dump H* hex
	Event Log 0 "Error while handing $handler:$reason\n$hex"
}

# Status handling
handler ContactStatus {uin status {ip 0}} {
	pset Contact:ICQ:$uin|Status $status
}

handler ClientInfo {uin args} {
	pset Contact:ICQ:$uin|Client $args
}

handler _ClientInfo {uin ver client unicode} {
	set uid Contact:ICQ:$uin
	pset $uid|Client "$client, ICQv${ver}\n[expr ${unicode}?{Unicode}:{Encoding: [pget $uid|encoding]}]"	
}

handler MyStatus {status} {
	pset Global:ICQ:Info|Status $status
	if {$status=="offline"} {
		foreach id [lsprop Contact:ICQ:*|Alias] {
			pset [lindex [split $id |] 0]|Status offline
		}
	}
}

# Translate ICQ message to uniform message
handler IncomingMessage {uin time message} {
	Event Incoming text Contact:ICQ:$uin $time $message
}

handler PagerMessage {uin time message} {
	foreach {nick a b mail cnt msg} $message break
	Event Incoming text Contact:pager:0 $time "Nick: $nick\nMail: $mail\n${msg}"
}

# Translate ICQ URL message to uniform message
handler IncomingURL {uin time message} {
	Event Incoming {descr URL} Contact:ICQ:$uin $time $message
}

handler Contacts {uin time message} {
	Event IncomingContacts Contact:ICQ:$uin $message 
}

handler Outgoing {type uid message} {
	if {![string match Contact:ICQ:* $uid]} return
	set uin [lindex [split $uid :] 2]
	switch -- $type {
		text {icq::message Message $uin $message}
	}
} ZZ_outgoing_to_wire


proc CreateProxy {} {
	package require proxy
	set p Global:ICQ:Proxy
	set proxy [proxy::proxy -type [pget $p|type] -server [pget $p|server]\
		-port [pget $p|port]]
	if {[pget $p|user]!={}} { 
		$proxy configure -user [pget $p|user]
		$proxy configure -password [pget $p|password]
	}
	set proxy
}

handler {Global:ICQ:Info|* Global:ICQ:Network|*} {id oldval newval} {
	array set binding {
		Global:ICQ:Info|Uin -uin
		Global:ICQ:Info|Password -password
		Global:ICQ:Info|Encoding -encoding
		Global:ICQ:Network|server -server
		Global:ICQ:Network|port -port
		Global:ICQ:Network|reconnect -reconnect
		Global:ICQ:Network|StartupStatus -status
	}
	if {[info exists binding($id)]} {
		icq::configure $binding($id) $newval
	}
}

handler Global:ICQ:Proxy|* {id oldval newval} {
	variable proxy 
	set prop -[lindex [split $id |] 1]
	if {$prop=={-type}} {
	   if {$newval=={none}} {
	   	icq::configure -proxy {}
	   } else {
	   	if {$proxy=={}} {set proxy [CreateProxy]}
	   	icq::configure -proxy $proxy
	   }
	} else { 
		if {$proxy!={}} { $proxy configure $prop $newval }
	}
}

proc PreInit {} {
	icq::icq -event ::Event -keepalive yes
}

proc PostInit {} {
	icq::configure -encoding [pget Global:ICQ:Info|Encoding]
	# Set alias to UIN, if unspecified
	if {![llength [lsprop Global:ICQ:Info|Alias]]} { 
		pset Global:ICQ:Info|Alias [pget Global:ICQ:Info|Uin]
		handler MyStatus {status} {
			if {$status=={offline}} return
			set ref [expr [clock seconds] & 0xff]
			hook ContactInfo "ContactInfo Global:ICQ:Info $ref"
			icq::ShortInfoRequest [pget Global:ICQ:Info|Uin] $ref
			unhook MyStatus myAlias
		} myAlias 
	}
	# Register all ICQ contacts
	icq::AddToList contacts [map x [lsprop Contact:ICQ:*|Alias] {
		lindex [split [lindex [split $x |] 0] :] 2} ]
	# Register ICQ Visible and Invisible lists
	foreach id {visible invisible} {
		icq::AddToList $id\
	      		[map y [grep x [lsprop Contact:ICQ:*|$id] {
				[string is true [pget $x]]}
	      		] {lindex [split [lindex [split $y |] 0] :] 2}]
	}
	handler Contact:ICQ:*|*visible {id oldval newval} {
		foreach {uid prop} [split $id |] break
		set cmd [expr {[string is true $newval]?{AddTo}:{RemoveFrom}}]
		icq::${cmd}List $prop [lindex [split $uid :] 2]
	}
	icq::start
}

PreInit

property	Global:ICQ:Info|Uin		{-type integer -save change}
property	Global:ICQ:Info|Password	{-type password -save change}
property	Global:ICQ:Info|Status		{-type variant -default offline -values {offline online occ dnd ffc away na invisible}}
property	Global:ICQ:Network|server	{-type string -default login.icq.com -save change}
property	Global:ICQ:Network|port		{-type string -default 5190 -save change}
property	Global:ICQ:Network|reconnect	{-type boolean -default 0 -save change}
property	Global:ICQ:Network|StartupStatus {-type variant -save change -default offline -values {offline online occ dnd ffc away na invisible}}

property	Global:ICQ:Proxy|type		{-type variant -default none -values {none http} -name {Network:Proxy:Type} -weight 02:01:01 save change}
property	Global:ICQ:Proxy|server		{-type string -default {} -save change}
property	Global:ICQ:Proxy|port		{-type integer -default {} -save change}
property	Global:ICQ:Proxy|user		{-type string -default {} -save change}
property	Global:ICQ:Proxy|password	{-type password -default {} -save change}

property	Contact:ICQ:*|Status		{-type variant -default offline sChanged -values {offline online occ dnd ffc away na invisible}}
property	Contact:ICQ:*|Client		{-type list -default {}}

property::Add	Contact:ICQ:*|visible	{-type boolean -weight 02:01 -name {Lists:Visible} -save change -default 0 }
property::Add	Contact:ICQ:*|invisible	{-type boolean -weight 02:01 -name {Lists:Invisible} -save change -default 0 }

set weight 1
foreach type {Message URL SMS Authorization AuthReq Contacts} {
	property Contact:ICQ:*|Send:$type [list -type action -weight 01:$weight -name Send:$type]
	incr weight
}
unset weight

after idle [nc PostInit]

