proc ChangeRoster {uid} {
	set action add
	if {[pget $uid|rid]} { set action update}
	if {[nil? [pget $uid|Groups]]} { set action delete }
	icq::roster $action [Roster:New $uid]
}

proc Roster:New {uid} {
	if {![pget $uid|rid]} {pset $uid|rid [rid]}
	switch -glob -- $uid {
		Contact:ICQ:* {
			set type contact
			set name [lindex [split $uid :] end]
			set grp [lindex [pget $uid|Groups] 0]
			if {[nil? $grp]} {set grp 0
			} else { set grp [pget $grp|rid]} 
			set rid [pget $uid|rid]
			set aux [list Alias [pget $uid|Alias]]
		}
		Group:common:* {
			set type group
			set name [pget $uid|Alias]
			set id [lindex [split $uid :] end]
			if {[string is integer $id] && $id<32768 && $id>0} {
				set grp $id
			} else {set grp [pget $uid|rid]}
			set rid 0
			set aux [list]
		}
		default { return }
	}
	list $type $name $grp $rid $aux
}

handler Roster {items time} {
	foreach x $items {
		foreach {type name group id aux} $x break
		Event Log 3 "roster: $type, $name, $group, $id: {$aux}"
		if {[llength [info procs Roster:$type]]} {
			Roster:$type $name $group $id $aux
		}
	}
	Event Log 3 "Roster updated: [clock format $time -format {%x %X}]"
	handler {Contact:ICQ:*|Alias Contact:ICQ:*|Groups
		Group:common:*|Alias} {id old new} {
			delayed [namespace current]::ChangeRoster [lindex [split $id |] 0]
	} rChanged 
	Roster:Sync
}

proc Roster:group {name group id aux} {
	if {!$group} return
	set uid Group:common:$group
	# Check if we already have mapped group
	foreach x [lsprop Group:common:*|rid] {
		if {[pget $x]==$group} {
			set uid [lindex [split $x |] 0]
			break
		}
	}
	# If mapped group not found, try to make mapping by name
	if {![llength [lsprop $uid|Alias]]} {
		foreach x [lsprop Group:common:*|Alias] {
			if {[pget $x]==$name} {
				set uid [lindex [split $x |] 0]
			}
		}
	}
	# Create or modify group
	if {[pget $uid|Alias]=={unknown}} { pset $uid|Alias $name }
	if {[nil? [pget $uid|Groups]]} { pset $uid|Groups {} }
	pset $uid|rid $group
}

proc Roster:contact {name group id aux} {	
	set uid Contact:ICQ:$name
	# set alias if not set yet
	array set Aux $aux
	if {[pget $uid|Alias]=={unknown}} { pset $uid|Alias $Aux(Alias) }
	# Make group mapping
	set grp Group:common:$group
	foreach x [lsprop Group:common:*|rid] {
		if {[pget $x]==$group} {
			set grp [lindex [split $x |] 0]
		}
	}
	# Add contact to group only if it is not a member of the group yet
	set groups [pget $uid|Groups]
	if {[nil? $groups]} {set groups [list]}
	if {[lsearch $groups $grp]==-1} {
		lappend groups $grp
		pset $uid|Groups $groups
	}
	# Set random id for contact
	pset $uid|rid $id
	icq::AddToList contacts [list $name]
}

proc Roster:Sync {} {
	Event Log 0 "Roster sync"
	set roster [list]
	foreach x [concat [lsprop Group:common:*|Groups] [lsprop Contact:ICQ:*|Groups]] {
		set uid [lindex [split $x |] 0]
		if {![pget $uid|rid]} {
			pset $uid|rid [rid]
			lappend roster add [Roster:New $uid]
		}
	}
	if {[llength $roster]} {
		Event Log 3 "Adding: $roster"
		eval icq::roster $roster
	}
}

handler Global:ICQ:Roster|Use {id old new} {
	icq::configure -roster $new
}

proc rid {} {
	set rids [map x [lsprop *|rid] {pget $x}]
	while 1 {
		set rid [expr {int(rand()*32000.0+500)}]
		if {[lsearch $rids $rid]==-1} {return $rid}
	}
}

property::Add	Contact:ICQ:*|rid	{-type integer -default 0 -save change}
property::Add	Group:common:*|rid	{-type integer -default 0 -save change}
property	Global:ICQ:Roster|Use		{-type boolean -default 0 -save change}
property	Global:ICQ:Roster|Modified	{-type integer -default 0 -save change}

