# This module is part of Alicq package. It is used for saving values of 
# properties changed by user when working with Alicq
#
# Author: Ihar Viarheichyk

# Save only properties which have -save flag set and value is different from 
# default
proc SaverFilter {mask} {
	grep x [lsprop $mask] {
		[property::exists $x -save]&&\
	     	![string equal [pget $x] [property::meta $x -default]]
	}
}

# Save properties we know how ho save
proc ItemSaver {id oldval newval} {
	foreach {uid prop} [split $id |] break
	set file $::RCFILE
	foreach {key file} [pget Global:Saver|paths] {
		if {[string match $key $uid]} break
	}
	switch -glob -- $id {
		Group:common:* -
		Contact:ICQ:* {
			foreach {type _ id} [split $uid :] break
			set key "^\\s*${type}\\s+${id}\\s"
			set grp [pget $uid|Groups]
			if {[nil? $grp]} {property::Save $file $key [nil]; return}
			set value "$type ${id} {[pget $uid|Alias]}"
			set grp [map g $grp {lindex [split $g :] 2}]
			append value " {$grp}"
			set app [join [map x [grep y [SaverFilter $uid|*] {
				![string match *|Alias $y]&&\
				![string match *|Groups $y]
				}] {
				set v [lindex [split $x |] 1]=[pget $x]
				expr {([string first { } $v]!=-1)?"{$v}":$v}
				}] { }]
			if {$app!={}} {append value " {$app}"}	
			property::Save $file $key $value
		}
		Global:* {
			regsub {[|]} $id {\\&} rid
			set key "^\\s*pset\\s+$rid"
			set val $newval
			if {[string first { } $val]!=-1} {set val "{$val}"}
			set val "pset $id $val"
			property::Save $file $key $val
			return
		}
		* {puts "Don't know how to save $id"; return}
	}
}

# Register hook handler for all properties which have save key
# set to 'change'
proc RegisterForSave {args} {
	variable forsave
	unhook * ItemSaver
	foreach prop [lsmeta] {
		if {[property::exists $prop -save] &&\
		    [property::meta $prop -save]=="change"} {
			hook $prop ItemSaver
		}
	}
	foreach prop [lsprop] {
		if {[property::exists $prop -save] &&\
		    [property::meta $prop -save]=="exit"} {
			set forsave($prop) [pget $prop]
		}
	}
}

handler ModuleLoaded {args} { delayed [nc RegisterForSave] }

# Save all properties which have key save set to 'exit' and changed 
# their values
handler Global|Exit {args} {
	variable forsave
	foreach prop [lsprop] {
		if {[property::exists $prop -save] &&\
	    	    [string equal [property::meta $prop -save] exit]} {
		    	if {[info exists forsave($prop)]} {
				set val $forsave($prop)
			} else {
				set val [property::meta $prop -default]
			}
		    	if {[pget $prop]!=$val} {
		    		ItemSaver $prop _ [pget $prop]
			}	
		}
	}
}

property::Add Global:Saver|paths {-type list -default {}}
