proc Grep {} {
	map y [grep x [lsmeta] {
		[string match Global* $x] &&\
		![string equal [property::meta $x -type] action]
	}] {lindex [split $y |] 0}
}

proc PopulateTree {tree item} {
	set parent root 
	if {[llength $item]>1} {
		set parent [PopulateTree $tree [lrange $item 0 end-1]]
	}
	set node [join $item :]
	if {![$tree exists $node]} {
		$tree insert end $parent $node -text [lindex $item end]\
			-image img:group -open 1
	}	
	set node
}

proc Apply {pref} {
     # Check for password match	
     foreach {pw} [info vars ${pref}:check:*] {
	regsub {:check:} $pw {} var
	if {[set $pw]!=[set $var]} { 
		Event Error 0:80 "Password does not match"
		return 0
	}	
     }
     foreach var [info vars ${pref}*] {     
	if {![string match ${pref}:check:* $var]} {
		pset [string range $var [string length $pref] end] [set $var]
	}	
     }
     return 1
}

proc Clear {pref} {
     foreach var [info vars ${pref}*] { unset $var }
}

proc Select {tree frame pref node} {
	set curr [lindex [$tree selection get] 0]
	if {$curr!=$node} {
		$tree selection set $node
		ConfigPane $frame [grep x [lsmeta] {[string match $node|* $x]}] $pref
	}
}

proc ConfigPane {f props pref} {
	foreach slave [grid slaves $f] {destroy $slave}
	foreach p $props {
		if {![info exists $pref$p]} {
			set $pref$p [pget $p]
		}	
		switch -exact -- [property::meta $p -type] {
			boolean {
				set ctl [checkbutton $f.cb$p -variable $pref$p]
			}
			password {
				if {![info exists $pref:check:$p]} {
					set $pref:check:$p [set $pref$p]
				}	
				set ctl [ui::frame::password $f.p$p $pref$p ${pref}:check:$p] }
			integer -
			string {
				set ctl [entry $f.en$p -textvariable $pref$p]
			}
			variant {
				set ctl [ui::frame::variant $f.v$p $pref$p [property::meta $p -values]]
			}

		}
		if [info exists ctl] {
			label $f.lb$p -text [lindex [split $p |] 1]
			grid $f.lb$p $ctl -sticky we
			unset ctl
		}	
	}
}
handler Global|Configure {args} {
	set top .conf
	if {[catch {toplevel $top -class AlicqConfig}]} return
	wm geometry $top 500x350
	wm minsize $top 400 300
	pack [set pw [PanedWindow $top.pw -side top]] -expand yes -fill both
	set sw [ScrolledWindow [$pw add].sw -relief flat -borderwidth 1]
	Tree $top.tr -width 20
	$sw setwidget $top.tr
	set pane [frame [$pw add].fr]
	pack $sw $pane -expand yes -fill both
	frame $pane.fr -width 300 -height 200 -borderwidth 1 -relief sunken
	
	set pref [namespace current]::tmp_
	set btn [frame $pane.btn -relief sunken -borderwidth 1]
	button $btn.ok -text "OK" -command "if {\[[namespace current]::Apply $pref\]} {destroy $top}"
	button $btn.apply -text "Apply" -command [nc Apply $pref]
	button $btn.cancel -text "Cancel" -command "destroy $top"
	bind AlicqConfig <Destroy> [nc Clear $pref]
	
	pack $btn.ok $btn.apply $btn.cancel -side left -fill x -padx 10 -pady 4
	$top.tr bindText <1> [nc Select $top.tr $pane.fr $pref]

	pack $pane.fr -expand yes -fill x -side top -padx 10 -pady 10
	pack $btn -side bottom
	foreach i [Grep] { PopulateTree $top.tr [split $i :]}
}


property::Add Global|Configure {-type action -weight 70 -name Configure weight 94}
