# This module is a part of Alicq instant messenger package
# It provides core of basic graphical interface of Alicq: tree window
#
# Author: Ihar Viarheichyk

package require BWidget

namespace eval ::tree {

# Node properties
property Contact:*|Node:Image	[list -type string -default offline]
property Group:*|Node:Image	[list -type string -default group]
property *|Node:Text		{-type string -default ? }
property *|mapping		{-type list -default {} }
property *|groups		{-type list -default {root} }
property *|children		{-type list -default {} }
property Group:*|Open		{-type boolean -default 0 -save exit}
property Contact:*|Open		{-type boolean -default 0 }

property Global:UI:Tree|Grouping  {-type variant -values {Plain} -save exit
				   -default Plain -name Interface:Grouping}
property Global:UI:Tree|Sort:Groups  {-type variant -values {Unsorted}
	-default Unsorted -name {Interface:Sort groups by...} -save exit}
property Global:UI:Tree|Sort:Contacts  {-type variant -values {Unsorted}
	-default Unsorted -name {Interface:Sort contacts by...} -save exit}
# Aux properties
property::Add	*|Cancel  {-type action -name Cancel -weight 99}


# Main tree window
proc MainWindow {} {
	wm group . .
	# Handle geometry command-line option
	if {![info exists ::geometry] &&
	    [set gm [option get . geometry Geometry]]!=""} {wm geometry . $gm }
	# Handle state of main window on startup: normal or iconic	 
	if {[set state [option get . state State]]!=""} { wm state . $state }
	# Create scrolled tree widget
	set sw [ScrolledWindow .sw -relief flat -borderwidth 1]
	set tree [Tree .tr -relief sunken -deltay 18\
		-dragenabled yes -dropenabled yes -dropcmd [nc DropCmd]\
		-opencmd [nc OpenCmd 1] -closecmd [nc OpenCmd 0]]
	$sw setwidget $tree
	::baloon::create
	foreach id {Text Image} {
		$tree bind${id} <<Item-Activate>> [nc DoubleClick]
		$tree bind${id} <<Context-Menu>> [nc PopupMenu %X %Y]
		$tree bind${id} <Enter> [nc DelayedHint 700]
		$tree bind${id} <Leave> [nc CancelHint]
	}
	set menu1 [menubutton .main -text "Menu" -menu .main.m]
	menu .main.m -type normal -tearoff no
	set menu2 [menubutton .status -text "Offline" -menu .status.m -image img:offline]
	grid $sw -columnspan 2 -sticky news
	grid $menu2 $menu1 -sticky w
	grid columnconfigure . 1 -weight 1
	grid rowconfigure . 0 -weight 1

	# Small hack to allow mouse wheel scrolling in tree widget
	bind .tr.c <4> ".tr.c yview scroll -2 units"
	bind .tr.c <5> ".tr.c yview scroll 2 units"
	
	wm protocol . WM_DELETE_WINDOW {Event Global|Exit}
	bind . <<Item-Activate>> [namespace code  {DoubleClick [.tr selection get]; break}]
}

proc DelayedHint {delay node} {
	variable hintdelay
	if {[info exists hintdelay]} {after cancel $hintdelay}
	set hintdelay [after $delay [nc ShowHint $node]]
}

proc ShowHint {node} {
	set uid [UidByNode $node]
	foreach {class type id} [split $uid :] break
	set hint "$type id: $id"
	if {[string match Contact:ICQ:* $uid]} { 
		set info [pget $uid|Client]
		if {[llength $info]} {
			foreach {ver client unicode} $info break
			append hint "\n$client, ICQv${ver}\n"\
			  [expr ${unicode}?{Unicode}:{Encoding: [pget $uid|encoding]}]
		}
		append hint "\nStatus: [pget $uid|Status]"
	}
	::baloon::show $hint
}

proc CancelHint {node} {
	variable hintdelay
	after cancel $hintdelay 
}

proc OpenCmd {val node} { pset [UidByNode $node]|Open $val }

proc DoubleClick {node} {
	if {[set uid [UidByNode $node]]=={}} return
	# If  node is group: open/close
	if [string match Group:common:* $uid] {
		pset $uid|Open [expr [pget $uid|Open]^1]
	} else { Event $uid|Send:Message $uid|Send:Message [clock seconds] }
}

proc CreateMenus {} {
	menu .status.m -type normal -tearoff no
	foreach status {online away dnd occ ffc na invisible offline} {
		if {$::tk_version >= 8.4} {
			.status.m add command -image img:$status\
				-label $status -compound left\
				-command "icq::SetStatus $status"
		} else {	
			.status.m add command -image img:$status\
				-command "icq::SetStatus $status"
		}
		pset Group:status:$status|Alias $status
		pset Group:status:$status|groups {}
	}
}

# --- Handle in-tree item renaming
handler {*|Rename} {id} {
	variable CurrentNode
	set uid [lindex [split $id |] 0]
	.tr edit $CurrentNode [pget $uid|Alias] [nc ItemModify $uid]
}

proc ItemModify {uid text} {
	pset $uid|Alias $text
	return 1
}

# --- Handle moving items between groups
proc DropCmd {dest src type op datatype data} {
	# Don't accept unknown types and don'w allow loops
	if {![string match node* $type]||\
	     [string match ${data}* [lindex $type 1]]} return
	set dest [UidByNode [lindex $type 1]]
	switch -glob -- $dest {
		Group:common:* {
			set uid [UidByNode $data]
			if {$uid==$dest} return
			set old [UidByNode [NodeParent $data]]
			# Set filter expression depending of drop type
			if {$op!={copy}} {set exp {$x!=$old}} else {set exp 1}
			append exp {&& $x!=$dest}
			pset $uid|Groups [concat\
				[grep x [pget $uid|Groups] $exp] $dest]
		}
	}
}

# --- Handle deletion of an item
handler {*|Delete} {id} {
	variable CurrentNode
	set uid [UidByNode $CurrentNode]
	set parent [UidByNode [NodeParent $CurrentNode]]
	pset $uid|groups [grep x [pget $uid|groups] {$x!=$parent}]
	if {![llength [pget $uid|groups]]} { RealDelete $uid 
	} else {
		pset $uid|Groups [grep x [pget $uid|Groups] {$x!=$parent}]
	}
}

# --- Real deletion from contact list
proc RealDelete {uid} { 
	pset $uid|Groups [nil]
	foreach x [lsprop $uid|*] { property::Unset $x}
}

# --- Context menu handler
proc PopupMenu {x y node} {
	variable CurrentNode
	Event CreateContextMenu [UidByNode $node] .context
	set CurrentNode $node
	.context post $x $y
}

handler ConfigLoaded {args} {
	property::meta Global:UI:Tree|Grouping -values\
		[map x [info procs Grouping::*] { namespace tail $x }]
	property::meta Global:UI:Tree|Sort:Contacts -values\
		[map x [info procs sort::Contact::*] { namespace tail $x }]
	property::meta Global:UI:Tree|Sort:Groups -values\
		[map x [info procs sort::Group::*] { namespace tail $x }]
	if {![llength [lsprop Group:common:other|*]]} { Group other Other }
}

proc FlashIcon {uid stage} {
	pset $uid|groups [Grouping::[pget Global:UI:Tree|Grouping] $uid]
	pset $uid|Node:Image [expr {$stage?[pget $uid|Status]:{message}}]
	set stage
}

# --- Flashing icon when pending messages present
handler Contact:*|Pending {id oldval newval} {
	if {[llength $oldval]&&[llength $newval]||[nil? $newval]} return
	set uid [lindex [split $id |] 0]
	set cmd "FlashIcon $uid"
	if {[llength $newval]>0} {
		hook Flashing $cmd
		Flash more
	} else {
		unhook Flashing $cmd
		FlashIcon $uid 1
		Flash less
		pset $uid|groups [Grouper $uid]
	}
}

# Namespace containing different grouping policies
namespace eval Grouping {
	variable Groups {*|Groups}
	variable Plain {*|Groups}
	variable Status {*|Groups Contact:*|Status}
	variable Online {*|Groups Contact:*|Status}

	# No grouping (default)
	proc Plain {uid} {
		expr {[string match Contact:* $uid]?{root}:{}}
	}
	# Use grous
	proc Groups {uid} {
		set g [pget $uid|Groups]
		expr {[llength $g]?$g:{root}}
	}
	# Group by status
	proc Status {uid} {
		switch -glob -- $uid {
			Contact:* { return Group:status:[pget $uid|Status]}
			Group:status:* {return root}
			* {return {}}
		}
	}
	# Divide into two groups: online and offline contacts
	proc Online {uid} {
		switch -glob -- $uid {
			Contact:* {
			   return Group:status:[expr {([pget $uid|Status]==\
			   		{offline})?{offline}:{online}}]
			}
			Group:status:online {return root}
			Group:status:offline {return root}
			* {return {}}
		}
	}
}

proc filter {name body} {
	variable filters
	set filters($name) $body
	foreach x [lsort [array names filters]] { append result $filters($x) }
	if {[llength [info procs Grouper]]} {rename Grouper {}}
	proc Grouper {uid} $result
}

# Grouping changes handling
handler Global:UI:Tree|Grouping {id old new} {
	unhook [set Grouping::$old] grouper
	filter 99Grouper [info body [namespace current]::Grouping::$new]
	# Handle changes of 'Groups' property in 'by Groups' grouping policy
	if {[info exists Grouping::$new]} {
		handler [set Grouping::$new] {id old new} {
			set uid [lindex [split $id |] 0]
			pset $uid|groups [Grouper $uid]
		} grouper 
	}
	foreach id [lsprop *|groups] {
		set uid [lindex [split $id |] 0]
		pset $id [Grouper $uid]
	}
}

# Handle Plain mode - remove lines
handler Global:UI:Tree|Grouping {id old new} {
	if {$new=={Plain}} {.tr configure -showlines 0 -deltax 0
	} elseif {$old=={Plain}} {
		.tr configure -showlines [option get .tr showlines D]\
			      -deltax [option get .tr deltax D]
	}
}

# Handle changes of 'group' property - remap item and it's children
handler {*|groups} {id oldval newval} {
	set uid [lindex [split $id |] 0]
	foreach {common newgrp oldgrp} [lcompare $oldval $newval] break
	foreach n $newgrp {lappend [property::Variable $n|children] $uid }
	foreach n $oldgrp {
		pset $n|children [grep x [pget $n|children] { $x!=$uid }]
	}
	pset $uid|mapping [Remap $uid]
}

# Calculate mapping of item to Tree node
proc Remap {uid} {
	set res [list]
	foreach g [pget $uid|groups] {
		foreach n [pget $g|mapping] { lappend res $n/$uid }
	}
	set res
}

handler {Contact:*|Alias Group:*|Alias} {id oldval newval} {
	set uid [lindex [split $id |] 0]
	pset $uid|Node:Text $newval
	DoSort $uid
}

handler Contact:ICQ:*|Status {id oldval newval} {
	pset [lindex [split $id |] 0]|Node:Image $newval
}

handler {Global:ICQ:Info|Alias Global:ICQ:Info|Status} {args} {
	set status [pget Global:ICQ:Info|Status]
	wm title . "Alicq: [pget Global:ICQ:Info|Alias] $status"
	.status configure -image img:$status
}

handler {*|Node:* Group:*|Open} {id oldval newval} {
	foreach {uid prop} [split $id |] break
	foreach node [pget $uid|mapping] {
		switch -glob -- $prop {
			Node:Text {.tr itemconfigure $node -text $newval}
			Node:Image {.tr itemconfigure $node -image img:$newval}
			Node:Color {.tr itemconfigure $node -fill $newval}
			Node:Font {.tr itemconfigure $node -font $newval}
			Open {.tr itemconfigure $node -open [string is true $newval]}
		}
		#delayed [nc sort [NodeParent $node]]
	}
}

proc CreateNode {node} {
	set uid [UidByNode $node]
	.tr insert end [NodeParent $node] $node -text [pget $uid|Node:Text]\
		-image img:[pget $uid|Node:Image] -open [pget $uid|Open]
	delayed [nc sort [NodeParent $node]]
}

proc RemoveNode {node} { 
	.tr selection remove $node
	.tr delete $node 
}

proc UidByNode {node} { lindex [split $node /] end }

proc NodeParent {node} {string range $node 0 [expr [string last / $node]-1]}

# Sort namespace includes all available sorting methods
namespace export UidByNode
namespace eval sort {
	namespace export *
	namespace import [namespace parent]::UidByNode
	# No sorting at all
	proc Unsorted {s1 s2} {return 0}
	# Sort by identifier (e.g UIN for ICQ contacts, group ID for groups)
	proc Identifier {s1 s2} {
		string compare [UidByNode $s1] [UidByNode $s2]
	}
	# Sort by alphabet (case-sensitive)
	proc Alphabet {s1 s2} {
		string compare	[pget [UidByNode $s1]|Node:Text]\
				[pget [UidByNode $s2]|Node:Text]
	}
	# Sort by alphabet (case-insensitive)
	proc {Alphabet,case-insensitive} {s1 s2} {
		string compare -nocase [pget [UidByNode $s1]|Node:Text]\
				       [pget [UidByNode $s2]|Node:Text]
	}
	# Sort by status and then by alphabet
	array set weight {ffc a online b occ c dnd d away e
			na f invisible g offline z}
	proc Status {s1 s2} {
		variable weight
		set st1 [pget [UidByNode $s1]|Status]
		set st2 [pget [UidByNode $s2]|Status]
		set res [string compare $weight($st1) $weight($st2)]
		if {$res} {set res} else { Alphabet $s1 $s2 }
	}
	# Contact-specific sorting
	namespace eval Contact {
		namespace import [namespace parent]::*
		namespace forget UidByNode
	}
	#Group-specific sorting
	namespace eval Group {
		namespace import [namespace parent]::*
		namespace forget UidByNode Status
	}
}

proc comparator {n1 n2} {
	foreach x {1 2} {set type$x [lindex [split [UidByNode [set n$x]] :] 0]}
	if {[set res [string compare $type1 $type2]]} { return $res }
	sort::${type1}::[pget Global:UI:Tree|Sort:${type1}s] $n1 $n2
}

proc sort {x} {
	if {![.tr exists $x]} return
	.tr reorder $x [lsort -command comparator [.tr nodes $x]]
	# workaround for nonupdating tree in Plain mode
	if {$x=={root}} {.tr insert end root _fake_; .tr delete _fake_}
}

handler Global:UI:Tree|Sort:* {id old new} {
	foreach m [concat root|mapping [lsprop Group:*|mapping]] { 
		foreach node [pget $m] {delayed [nc sort $node]}
	}
	if {$new=={Status}} { 
		handler Contact:*|Status {id args} {DoSort $id} 99sort
	}
	if {$old=={Status}} { unhook Contact:*|Status 99sort }
}

proc DoSort {id} {
	foreach x [pget [lindex [split $id |] 0]|mapping] { 
		delayed [nc sort [NodeParent $x]]
	}
}

pset root|mapping root
MainWindow
#Handle tree nodes mapping changes
handler {*|mapping} {id oldval newval} {
	foreach {_ Create Remove} [lcompare $oldval $newval] break
	foreach a {Create Remove} { foreach n [set $a] { ${a}Node $n } }
	set uid [lindex [split $id |] 0]
	foreach n [pget $uid|children] { pset $n|mapping [Remap $n]}
}


CreateMenus
pset Global:UI:Tree|Grouping Plain
}
