# This package is used to convert old (versions 0.1-0.6.5) alicq startup file
# to new format, used in alicq 0.8
# It can also be used for repairing damaged config file in future
#
# Copyright Ihar Viarheichyk (iverg@mail.ru) 2002

package provide repair 0.8

# Try gently recover config - replace wrong statements only
proc gentle {config} {
}

# Try to recover everything we can, if gentle recovering failed. Config file
# will be completely rewritten.
proc force {config} {
}

proc convert {config} {
	set substs {
		Uin			Global:ICQ:Info|Uin
		Password		Global:ICQ:Info|Password
		Alias			Global:ICQ:Info|Alias

		port			Global:ICQ:Network|port
		server			Global:ICQ:Network|server
		reconnect		Global:ICQ:Network|reconnect
		status			Global:ICQ:Network|StartupStatus

		proxy_server		Global:ICQ:Proxy|server
		proxy_port		Global:ICQ:Proxy|port
		proxy_user		Global:ICQ:Proxy|user
		proxy_password		Global:ICQ:Proxy|password
		
		base::PopupDialog	Global:UI:Messages|Popup
		base::HideOffline	Global:UI:Tree|HideOffline
		base::HideEmptyGroups	Global:UI:Tree|HideEmpty
		base::GroupingRule	Global:UI:Tree|Grouping
		base::SortingRule	Global:UI:Tree|Sort

		url::handler\\(([\\w]*)\\) Global:UI:Url|\\1-handler
		url::regexp\\(([\\w]*)\\) Global:UI:Url|\\1-regexp
	}
	foreach {exp subst} $substs {
		regsub -all -- "set\\s+$exp\\s" $config "pset $subst " config
	}
	regsub -all -- {Module\s+(GUI|modules)/} $config {Module } config
	regsub -line -all -- {^\s*set\s+base::(Group|Contact)sSortingRule.*?(;|$)} $config {} config
	foreach {r1 r2} {plain Plain groups Groups status Status onoff Online/Offline} {
		regsub "\\|Grouping\\s+$r1" $config "|Grouping $r2" config
	}
	regsub -line -all -- {^set\s+(base)::.*(;$)} $config {} config
	if {![regexp -- {Module\s+icqconn} $config]} {
		set config "Module icqconn.tcl\n$config"
	}
	return $config
}

proc detect {config} {
	set signs {
	  {^\s*set\s+base::} {^\s*set\s+(Uin|Alias|Password)}
	}
	foreach sign $signs {
		if {[regexp -line -- $sign $config]} {return 1}
	}
	return 0
}
# If called from Alicq itself, work with RCFILE, otherwise with stdin/stdout
proc proceed {} {
	if {[info exists ::RCFILE]} {
		set fd [open $::RCFILE]
		set config [read $fd]
		close $fd
	} else {
		set config [read stdin]
	}

	if {1 || [detect $config]} {
	    set config [convert $config]
	    if {[info exists ::RCFILE]} {
		set backup ${::RCFILE}.old
		if {![file exists $backup]} { file copy $::RCFILE $backup }
		set fd [open $::RCFILE w]
		puts $fd $config
		close $fd
	    } else {
	      puts $config
	    }
	}
}
if {[info script]==$argv0 || [info exists ::RCFILE]} { proceed } 

