#  ICQ client library
#  Version 0.2	
#  Copyright (C) Ihar Viarheichyk 2001

#  This library gives ability to use ICQ v8 protocol (OSCAR) in tcl programs.
#  Protocol description and some ideas in implementattion were taken from 
#  ICQ2000.pm and ICQ2000_Easy.pm perl modules by
#  Robin Fisher <robin@phase3solutions.com> and vICQ program by 
#  Alexander Timoshenko

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
	       
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

package provide icq 0.8.5
namespace eval icq {
	variable SeqNum	3412
	variable rid 0x312
	variable delim [binary format c 0xFE]
	variable glue {<...>}
	variable Defaults {	
		-server		login.icq.com	-port		5190
		-uin		0		-password	{}
		-encoding	{}		-register	{}
		-contactslist	{}		-visiblelist	{}
		-invisiblelist	{}		-reconnect	0
		-status		offline		-event		{empty}
		-proxy		{}		send_queue	{}
		Flags		0x0001		MyStatus	offline	
		ip		0.0.0.0		port		0
		-keepalive	0		keepalive	60000
		-unicode	0		-roster		0
	}

array set ErrorString { 24	"Connection rate exceeded" 
			5	"Wrong password" }

# Tlv format specification. Only non-string fields given
array set tlvSpec { 1:20 I 1:22 S 1:23 S 1:24 S 1:25 S 1:26 S 4:8 S
        2:1 S 2:2 I 2:3 I 2:4 S 2:6 I 2:8 S 2:10 I 2:15 I 2:18 S msg:10 S
	roster:202 c roster:200 S* roster:102 S}

# Alias table: map tlv, status etc codes to names
array set alias {
	1:1 -uin 1:2 -password 1:3 ClientProfile 1:5 UserInfo 1:6 Cookie
	1:14 Country 1:15 Language 1:20 BuildMin 1:22 ClientType
	1:23 VersionMaj 1:24 VersionMin 1:25 IcqNumber 1:26 BuildMaj
	2:1 UserClass 2:2 SignupDate 2:3 SignonDate 2:4 Port 2:5 EncodedMsg
	2:6 Status 2:8 ErrCode 2:10 IP 2:11 WebAddr 2:12 LanDetails
	2:13 Capabilities 2:15 TimeOnline 2:18 Unk1 4:1 UIN 4:4 HtmlAddr
	4:5 ServerAndPort 4:6 Cookie 4:8 ErrCode 4:9 DisconnectCode

	roster:202 visibility roster:305 Alias roster:200 ids roster:205 icqtic
	roster:200 members roster:102 auth
	rtype:0 contact rtype:1 group rtype:4 visibility rtype:9 icqtic
	raction:add 2:19:8 raction:update 2:19:9 raction:delete 2:19:10
	
	status:0 online status:1 away status:2 dnd status:4 na status:5 na
	status:16 occ status:17 occ status:19 dnd status:32 ffc 
	status:256 invisible

	cap:094613444c7f11d18222444553540000 ICQ
	cap:094613494c7f11d18222444553540000 ICQRELAY
	cap:0946134e4c7f11d18222444553540000 UTF8
	cap:97b12751243c4334ad22d6abf73f1492 RTF
	cap:a0e93f374c7f11d18222444553540000 STR_2001
	cap:10cf40d14c7f11d18222444553540000 STR_2002
	cap:2e7a6475fadf4dc8886fea3595fdb6df IS_2001
	cap:563fc8090b6f41bd9f79422609dfa2f3 WEB
	cap:f2e7c7f4fead4dfbb23536798bdf0000 TRILL_CRYPT
	cap:97b12751243c4334ad22d6abf73f1409 TRILL_UNK
	cap:97b12751243c4334ad22d6abf73f14?? Sim
	cap:6d49435120a920522e4b2e20???????? mICQ
	
	cli:ffffffff Miranda cli:ffffff8f StrICQ cli:ffffff42 mICQ	
	cli:ffffff7f &RQ cli:ffffffab YSM cli:ffffffbe Alicq

	Type:1 Message Type:4 URL Type:19 Contacts
	type:3 File type:4 {Send Web Page Address (URL)} type:19 Contacts
	type:2 {Send / Start ICQ Chat} 
}
# Add Alicq capability
set cap {Alicq:ver:}
foreach x [split [package present icq] .] {append cap [format %02x $x]}
#binary scan {Alicq, 2001-2003} H* alias(cap:ALICQ)
binary scan $cap H* alias(cap:Alicq)

# Reverse alias table
foreach {key val} [array get alias] { 
	foreach {tag id} [split $key :] break
	set alias($tag:$val) $id
}
#

proc log {level str} { $::icq::data(-event) Log $level $str }
proc empty {args} {}

proc icq {args} {
	upvar #0 icq::data data
	array set data $icq::Defaults
	set data(-encoding) [encoding system]
	eval configure $args
}

proc start {} {
	upvar #0 icq::data data
	if {$data(-status)!="offline"} { Connect icq::data $data(-server) $data(-port) }
}

proc configure {args} {
	upvar #0 icq::data data
	set options [lsort [array names data -*]]
	foreach {option value} $args {
		if {[lsearch -exact $options $option]==-1} {
			return -code error "Unknown option \"$option\""
		}
		if {$option=="-password"} {
			set value [EncryptPassword $value]
		}
		if {$option=="-proxy" && $value!={}} {
			$value configure -result [nc ProxyResult icq::data]
		}
		set data($option) $value
	}
}

proc SetEncoding {encoding {uin {}}} {
	upvar #0 icq::data data
	set data(${uin}-encoding) $encoding
}

proc AddToList {type Contacts} {
	upvar #0 icq::data data
	foreach contact $Contacts { lappend data(-${type}list) $contact }
	if {$data(MyStatus)!="offline"} {UploadList $type $data(-${type}list)}
}
proc RemoveFromList {type Contacts} {
	upvar #0 icq::data data
	foreach contact $Contacts {
		set pos [lsearch $data(-${type}list) $contact]
		set data(-${type}list) [lreplace $data(-${type}list) $pos $pos]
	}
	if {$data(MyStatus)!="offline"} {UploadList $type $data(-${type}list)}
}

proc ReadData {name} {
	upvar #0 $name data
	set chunk [read $data(socket) $data(Length)]
	append data(packet) $chunk
	if {$chunk=={}} {
		CloseConnection $name
		if {[string is true $data(-reconnect)]} {
		   after idle [nc Connect $name $data(-server) $data(-port)]
		}
	} elseif {![incr data(Length) -[string length $chunk]]} {
		if {$data(wantBody)} {
			ParsePacket $name $data(packet)
			set data(packet) {}
			set data(Length) 6
		} else {
			binary scan $data(packet) c@4S Id data(Length)
			if {$Id != 42} {log 0 "Unaligned data!"}
		}
		set data(wantBody) [expr $data(wantBody)^1]
	}
}

# Toplevel packet handler
proc ParsePacket {name packet} {
	upvar #0 $name data
	match $packet {{byte _ channel} 4 {* other}}
	set handler $channel:0:0 
	if {$channel!=1 && $channel!=4} { 
		match $other {{bword family subID} {* other}}
		set handler $channel:$family:$subID
	}
	if {[llength [info commands $handler]]} {
		if {[catch {$handler $name $other} reason]} {
			$data(-event) Error:Protocol $handler $reason $packet
		}
	} else { 
		binary scan $packet H* hex
		log 2 "There is no handler for message $handler\n$hex" 
	}
}
# These are handlers for server messages 

# Login negotiation
proc 1:0:0 {name packet} {
	variable loginTLV
	upvar #0 $name data
	log 2 "Got login invitation, logging in"
	if {$data(-register)!={}} { 
		Register $data(-register)
	} else {
		if {[info exists data(Cookie)]} {
			log 3 "Using cookie"
			set cmd [TLV 1 Cookie $data(Cookie)]
			unset data(Cookie) 
		} else { set cmd [TLV 1 -uin $data(-uin)\
				-password $data(-password)][set loginTLV]
		}
		Packet 1 [bdword 1]$cmd
	}
}
# Disconnect negotiation
proc 4:0:0 {name packet} {
	upvar #0 $name data
	variable ErrorString
	
	log 3 "Disconnect command"
	array set TLV [lindex [scanTLV $packet 4 4] 0]
	CloseConnection $name
	if { [info exists TLV(Cookie)] && [info exists TLV(ServerAndPort)] } {
		set data(Cookie) $TLV(Cookie)
		log 3 "Server wants us to reconnect to $TLV(ServerAndPort)"
		after idle icq::Connect $name [split $TLV(ServerAndPort) :]
	} elseif {[info exists TLV(ErrCode)]} {
		if {[info exists ErrorString($TLV(ErrCode))]} {
			 set error_string $ErrorString($TLV(ErrCode))
		} else { set error_string "Unknown" }
		$data(-event) Error 4:$TLV(ErrCode) $error_string
	} elseif {[string is true $data(-reconnect)]} { 
		after idle [nc Connect $name $data(-server) $data(-port)]
	} else { log 2 "stub: No disconnect handler here" }
}
# My status
proc 2:1:15 {name packet} {
	upvar #0 $name data 
	variable alias	

	match $packet {6 {buin User} {bword WarningLevel} {TLV tlvs}}
	array set TLV $tlvs
	
	if {$data(MyStatus)=="offline"}  { 
		if {[string is true $data(-keepalive)]} {
			KeepAliveStart $data(keepalive)
		}
		UploadList contacts $data(-contactslist) 
	}
	set data(MyStatus) $alias(status:[expr $TLV(Status)&255])
	if {$data(MyStatus)=="invisible"} { 
		UploadList visible $data(-visiblelist)
	}
	if {[string is true $data(-keepalive)]} RestartTimer
	$data(-event) MyStatus $data(MyStatus)
}

# Server Ready 
proc 2:1:3 {args} {
	Command 2:1:23 [bword 1 3 2 1 3 1 21 1 4 1 6 1 9 1 10 1]
}
# Rate info answer
proc 2:1:7 {args} {
	#Rate info ack
	Command 2:1:8 [bword 1 2 3 4 5]
	foreach cmd {2:1:14 2:2:2 2:3:2 2:4:4 2:9:2} { Command $cmd {}}
}

# Conact online
proc 2:3:11 {name packet args} {
	upvar #0 $name data 
	variable alias

	match $packet {6 {buin UIN} word {TLV tlvs}}
	array set TLV $tlvs
	if {[info exists TLV(Status)]} {
		set status [alias status [expr $TLV(Status)&65535]]
		if {[string is integer $status]} {
			log 0 "Unknonw status $status"
		} else { $data(-event) ContactStatus $UIN $status $TLV(IP) }
		if {$TLV(Status)&0x80000} { $data(-event) ICQBirthday $UIN }
		if {$TLV(Status)&0x20000} { $data(-event) ICQShowIp $UIN }
	}
	# Check capabilities
	if {[info exists TLV(Capabilities)]} {
		set data($UIN-unicode) 0
		set ver 0
		set lcaps [list]
		set str $TLV(Capabilities)
		while {$str!={}} {
			foreach {cap str} [cap.get $str] break
			lappend lcaps $cap
		}
		# Determine client
		foreach {mask client} [list [alias cap Sim] Sim WEB ICQ-Lite\
			UTF8 ICQ2002 *TRILL* Trillian IS_2001 ICQ2001\
			[alias cap mICQ] mICQ _ unknown] {
			 if {[lsearch $lcaps $mask]!=-1} break
		}
		if {[lsearch $lcaps ICQ]!=-1 && [lsearch $lcaps UTF8]!=-1} {
			set data($UIN-unicode) 1
		}
		if {$client=={ICQ-Lite} && [lsearch $lcaps ICQ]==-1} {
			set client ICQ2go
			#set data($UIN-unicode) 0
		} elseif {$client=={Sim}} {
			set p [lindex $lcaps [lsearch $lcaps [alias cap Sim]]]
			set p 0x[string range $p end-1 end]
			append client { } [expr $p/64-1].[expr $p&63]

		}
		log 6 "Caps: $UIN [join $lcaps ,]"
	}
	if [info exists TLV(LanDetails)] {
		binary scan $TLV(LanDetails) IIcSISSSSIII addr port type\
			ver a b c d e stamp1 stamp2 stamp3
		$data(-event) LanDetails $addr $port $type
		if {[expr { ($stamp1>>24)==0x7d }]} {
			set client Licq
			if {[expr {$stamp1&0x00800000}]} {append client "/SSL"}
			set v [expr $stamp1&0x007FFFFF]
			append client " [expr $v/1000].[expr $v%1000/10].[expr $v%10]"
		} else {
		   set x [format %x $stamp1]
		   if [info exists alias(cli:$x)] {set client $alias(cli:$x)}
		}	
		if {[info exists client] &&\
		    [lsearch {Alicq mICQ Miranda &RQ} $client]!=-1} {
			append client " [expr ${stamp2}>>24].[expr (${stamp2}>>16)&0xFF].[expr (${stamp2}>>8)&0xFF]"
			if {[set l [expr $stamp2&0xFF]]} { append client ".$l" }
		}
	}
	# Check signup and signon dates
	foreach x {SignupDate SignonDate} {
		if {[info exists TLV($x)]} { $data(-event) $x $UIN $TLV($x) }
	}
	if {[info exists client] && [info exists ver]} {
		$data(-event) ClientInfo $UIN $ver $client $data($UIN-unicode)
	}
}

# Contact offline
proc 2:3:12 {name packet args} {
	upvar #0 $name data
	match $packet {6 {buin UIN}}
	# Unicode is not supported in offline mode
	set data($UIN-unicode) 0
	$data(-event) ContactStatus $UIN offline
}

# Request rate info
proc 2:1:24 {name packet} {
	upvar #0 $name data
	#Set status
	SetStatus $data(-status) 0

	# Strange command (maybe remove it?)
	Command 2:1:17 [word 0 0]
	#Client ready
	Command 2:1:2 [byte 0 1 0 3 1 16 2 138 0 2 0 1 1 1 2 138 0 3 0 1\
			1 16 2 138 0 21 0 1 1 16 2 138 0 4 0 1 1 16 2 183 \
			0 6 0 1 1 16 2 183 0 9 0 1 1 16 2 183 0 10 0 1 1 16\
			2 183]
	Command 2:3:2 {}
	SendCaps
	# Offline messages request
	SrvMessage 60 {}
}

# Buddy request
proc 2:3:3 {name args} { 
	upvar #0 $name data
	# If server roster enabled, request list service rights
	if {[string is true $data(-roster)]} { Command 2:19:2 {} }
}

# List service granted, request roster from server
proc 2:19:3 {name args} { Command 2:19:4 {} }

# Reply roster
proc 2:19:6 {name packet} {
	upvar #0 $name data
	match $packet {6 byte {bword count} {* packet}}
	set roster [list]
	for {} {$count>0} {incr count -1} {
		match $packet {{pstr name} {bword gid uid type}
			{pstr tlv} {* packet}}
		lappend roster [list [alias rtype $type] $name\
			$gid $uid [lindex [scanTLV $tlv 100 roster] 0]]
	}
	match $packet {{bdword time}}
	$data(-event) Roster [encoding convertfrom utf-8 $roster] $time
	log 5 "Roster was updated: [clock format $time -format {%x %X}]"
}
# Roster update ACK
proc 2:19:14 {name packet} {
	match $packet {6 {word status}}
	log 3 "Byddy added with status $status"
}

# Roster OK
proc 2:19:15 {name packet} {
	upvar #0 $name data
	match $packet {{bdword timestamp} {lword size}}
	$data(-event) Roster:OK $timestamp $size
	log 3 "Roster ok. Size is $size, timestamp: $timestamp"
}

# Error message
proc 2:1:1 {name args} { [set ::${name}(-event)] Error 2:0 }

# Message Received
proc 2:4:7 {name packet} {
	match $packet {14 {bword Type} {buin Sender} word {TLV tlvs} {* other}}
	array set TLV $tlvs
	log 5 "Incoming message of type $Type"
	if {[info exists TLV(EncodedMsg)]} { 
		Incoming $name $Sender [clock seconds] 13\
			[query $name -encoding $Sender] $TLV(EncodedMsg)
		return
	} 
	match $other {{bword pref} {pstr rest} {* other}}
	if {$pref==4} {match $other {{bword $pref} {pstr rest}}}
	set cmd [lindex [info level 0] 0]:$Type
	if {[llength [info procs $cmd]]} { $cmd $name $rest $Sender
	} else {log 0 "Unknown message type: $Type"}
}

# Message Received - Type 1
proc 2:4:7:1 {name packet uin} {
	match $packet {word pstr word {pstr msg}}
	match $msg {{bword flags flags2} {* msg}}
	log 5 "Type-1 message flags: $flags/$flags2"
	if {$flags==2 && $flags2==0} {
		set enc unicode
		set msg [swab $msg]
	} elseif {$flags==3 && $flags2==0} { set enc iso8859-1
	} else {set enc [query $name -encoding $uin]}
	Incoming $name $uin [clock seconds] 1 $enc $msg
}

# Message Received - Type 4: service messages
proc 2:4:7:4 {name RawMessage uin} {
	match $RawMessage {4 {lword Sub} {lntsz msg} {* rest}}
	Incoming $name $uin [clock seconds] $Sub [query $name -encoding $uin] $msg $rest
}

# Message Received - Type 5: new ICQ messages
proc 2:4:7:2 {name packet uin} {
	variable alias
	match $packet {{word ack} {dword time rid} {* rest}}
	if {$ack} {
		log 5 "Got ack from $uin in 2:4:7:2"
		return
	} else {match $rest {{cap c} {* rest}} }	
	array set TLV [lindex [scanTLV $rest 100 msg] 0]
	if {[string length $TLV(10001)]==27} {
		log 3 "Strange message from $uin"
		Command 2:4:11 [dword $time $rid][bword 2][buin $uin][lword 3]$TLV(10001)
		return
	}
	match $TLV(10001) {{lnts aux1} {byte sign} 15 {byte type _} 
		{word _ _} {lntsz msg} {* other}}
	log 5 "Special type: $type, cap=[alias cap $c]"
	if {$type==26} {
		match $other {{lnts header} {dstr content}}
		match $header {16 {word mtype} {dstr descr}}
		match $content {{dstr msg}}
		if {[info exists alias(type:$descr)]} {
			set type $alias(type:$descr)
			Incoming $name $uin [clock seconds] $type utf-8 $msg
		} else {log 0 "Unknown extended message type: $descr"}	
	} elseif {$msg!={}} {
		set enc [query $name -encoding $uin]
		if {$type==1} { match $other {{dword fg bg} {* other}} }
		if {$other!={}} {match $other {{dstr guid}}
			match $guid {{guid c}}
			if {$c=={UTF8}} {set enc utf-8}
		}	
		Incoming $name $uin [clock seconds] $type $enc $msg
		Command 2:4:11 [dword $time $rid][bword 2][buin $uin][lword 3]$TLV(10001)
	} elseif {$sign==18} { log 3 "Some kind of service message" }
}

# Client acknowledgement
proc 2:4:11 {name packet} {
	upvar #0 $name data
	match $packet {10 {dword id} {bword type} {buin uin}}
	log 3 "Client ack from $uin: id=$id"
	$data(-event) ClientACK $uin $id
}

# Server Acknowledgement
proc 2:4:12 {name packet} {
	upvar #0 $name data
	match $packet {10 {dword id} {bword type} {buin uin}}
	log 3 "Server ack from $uin"
	$data(-event) ServerACK $uin $id
}


# All ICQ-specific stuff not fitting to original OSCAR
proc 2:21:3 {name packet} {
	match $packet {{bword _ Ref} 8 {ldword MyUIN} 
		{lword MessageType} {* RawMessage}}
	switch -exact -- $MessageType {
		65 {  match $RawMessage {{bword flags} {ldword UIN} {lword Y} 
	       	         {byte M D H Mi Type _} {lntsz msg} {* rest}}
		      Incoming $name $UIN [clock scan "$M/$D/$Y $H:$Mi"\
		      	-gmt 1] $Type [query $name -encoding $UIN] $msg $rest
		   }
		66 {  SrvMessage 62 {} 0
		      log 3 "Offline messages complete"
		   }
	      2010 { 2:21:3:2010 $name $RawMessage $MyUIN $Ref}	
	   default { binary scan $RawMessage H* hex
	   	     log 0 "UNKNOWN MESSAGE TYPE -----> $MessageType\n$hex"
		   }
	}
}

proc 2:21:3:2010 {name RawMessage MyUIN Ref} {
	upvar #0 $name data
	match $RawMessage {word {lword SubType} {byte result} {* RawMessage} }
	switch -glob -- $SubType:$result {
		*:50 { $data(-event) SearchResults $Ref {}}
		*:20 { $data(-event) SearchResults $Ref {}}
	      170:10 { $data(-event) PasswordChanged $Ref }
	      260:10 { log 3 "Short user info"
		       match $RawMessage {{lntsz i:Nick i:FirstName
			      i:LastName i:email}}
		     }
	      200:10 {	match $RawMessage {{lntsz i:Nick i:FirstName
			   i:LastName i:email i:City i:State i:Phone i:Fax 
			   i:Address i:Mobile i:Zip}
			   {lword i:Country} {byte i:TimeZone}}
		     }
	      235:10 {	match $RawMessage {{byte cnt} {* str}}
			for {set i 1} {$i<=$cnt} {incr i} {
				match $str [list byte [list lntsz i:email$i]\
			     			{* str}]
			}
			log 5 "Extra emails: $cnt"
		     }
	      210:10 {	log 3 "User info work" }	     
	      220:10 {	match $RawMessage {{lword i:Age} {byte i:Sex} 
				{lntsz i:Homepage} {lword i:Year} 
				{byte i:Month i:Day i:Lang1 i:Lang2 i:Lang3}}
		     }
	      230:10 {	log 3 "User info about" }
	      240:10 {	log 3 "Personal interests"}
	      250:10 {	log 3 "Past interests"}
	  4[123]0:10 {	match $RawMessage {{lnts packet}}
	  		match $packet {{ldword i:UIN} {lntsz i:Nick
			  i:FirstName i:LastName i:email} {byte i:AuthRq}
			  {lword i:Status}}
		     }
	      480:10 {	log 3 "User info unknown"}
	     2210:10 {	log 3 "XML" }
	     default {  log 1 "Unknown meta type $SubType:$result" }
	}
	set info [list]
	foreach x [info locals i:*] {
		lappend info [lindex [split $x :] 1] [FromWire [set $x] $data(-encoding)]
	}
	if {[llength $info]} {
		switch -glob -- [expr $SubType/100] {
		   2 { $data(-event) ContactInfo $Ref $info }
		   4 { $data(-event) SearchResults $Ref $info }
		}     
	}
	# Send empty result to indicate end of search
	if {$SubType==430} { $data(-event) SearchResults $Ref {} }
}

proc 2:23:5 {name packet} {
	binary scan $packet H* hex; log 0 "23:5:"; log 1 $hex
	CloseConnection $name
}

array set msgCodes {
	1 IncomingMessage 4 IncomingURL 5 AuthorizationRequest 19 Contacts
	6 AuthorizationRequest 7 AuthorizationDenied 8 AuthorizationGiven
	13 PagerMessage
}
proc Incoming {name uin time code enc msg { aux {} }} {
	upvar #0 $name data
	if {$code!=1} {
		 set msg [split $msg $icq::delim]
	} else { set msg [glue $name $msg $uin] }
	set msg [FromWire $msg $enc]
	if {[info exists icq::msgCodes($code)]} {
		if {$code==19} { set msg [lrange $msg 1 end-1] }
		log 5 "Message of subtype $code ($icq::msgCodes($code))"
		log 5 "Data: $msg"
		$data(-event) $icq::msgCodes($code) $uin $time $msg
	} else { log 0 "ERROR: unknown message code: $code" }
}

proc glue {name msg uin} {
	upvar #0 $name data
	set len [string length $icq::glue]
	if [string match ${icq::glue}* $msg] {
		set prev {}
		if [info exists data(inc:$uin)] {set prev $data(inc:$uin) }
		set msg ${prev}[string range $msg $len end]
	}
	if [string match *${icq::glue} $msg] {
		set data(inc:$uin) [string range $msg 0 end-$len]
		return -code return
	} 
	if [info exists data(inc:$uin)] { unset data(inc:$uin) }
	set msg 
}
# End of server messages handlers 

# Client commands
proc Register {password} {
	set sequence [byte 0 0 0 0 0x28 0 3 0 0 0 0 0 0 0 0 0 0x9e 0x27 0 0\
		0x9e 0x27 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0\
		0 0][lnts $password][byte 0x9e 0x27 0 0 0 0 0 0 3 2]
	Packet 1 [bdword 1]
	Command 2:23:4 [bword 1][pstr $sequence]
}

proc message {type uin val} {
	upvar #0 icq::data data
	if {$data(MyStatus)=="offline"} {
		$data(-event) Error 0:2 "Offline"
		return
	}
	$data(-event) Outgoing$type $uin [clock seconds] $val
	# Map type to corresponding code
	if {![string is integer $type]} {set type [alias Type $type]}
	if {![string is integer $type]} {
		return -code error "Unsupported message type $type"
	}
	set fmt [expr {[query icq::data -unicode $uin]?2:4}]
	set enc utf-8
	if {$fmt==4} {set enc [query icq::data -encoding $uin]}
	if {$fmt==4 && $type==1} {set fmt 1}
	# Add length and empty field for contacts list
	if {$type==19} { set val [concat [expr [llength $val]/2] $val {{}}] }
	set val [ToWire $val $enc]
	if {$type!=1} { set val [join $val $icq::delim] }
	message_raw $fmt $type $uin [binary format a* $val]
}

proc message_raw {fmt type uin val} {
	variable rid
	# Split big message to chunks
	if {$type==1 && [string length $val]>300} {
		after 2500 [nc message_raw $fmt $type $uin\
			${icq::glue}[string range $val 300 end]]
		set val [string range $val 0 299]${icq::glue}
	}
	Command 2:4:6 [dword 0 $rid][bword $fmt][buin $uin][message$fmt\
		$type $val][bword 6 0]
	set rid [expr {($rid+1)&0xFFFF}]
}

# Type 1: Simple message
proc message1 {type msg} { TLV msg 2 [byte 5 1 0 1 1 1 1][pstr [dword 0]$msg] }
# Type 4: Different kinds of messages
proc message4 {type msg} {
	upvar #0 icq::data data
	TLV msg 5 [ldword $data(-uin)][lword $type][lntsz $msg]]
}
# Type 2: Different kinds of messages, UTF-8
proc message2 {type msg} {
	variable rid
	set seq [lword 0xd1ff]
	set header [word 0][dword 0][ldword $rid]
	set aux [lnts [lword 8][dword 0 0 0 0][word 0][ldword 3][byte 0]$seq]
	append aux [lnts ${seq}[dword 0 0 0]][lword $type 0 0][lntsz $msg]
	if {$type==1} {append aux [bdword 0 0xffffff00]}
	append aux [dstr [guid UTF8]]
	TLV msg 5 ${header}[cap ICQRELAY][TLV msg 10 1 15 {} 10001 $aux]
}

proc roster {args} {
	if {![llength $args]} return
	# Start update of server-side roster
	Command 2:19:17 {} 
	# Proceed arguments as action-item pairs
	foreach {action x} [encoding convertto utf-8 $args] {
		foreach {type name group id aux} $x break
		log 0 "$action: $type/$name/$group/$id {$aux}"
		set cmd [pstr $name][bword $group $id [alias rtype\
			$type]][pstr [eval TLV roster $aux]]
		set c [alias raction $action]
		binary scan $cmd H* hex
		log 0 "Buddy $action: $hex"
		Command $c $cmd
	}
	# Update is finished
	Command 2:19:18 {} 
	after cancel {icq::Command 2:19:4 {}}
	after idle {icq::Command 2:19:4 {}}
}

proc Authorize {UIN {type 1}} {
	$::icq::data(-event) AuthorizationResponse $UIN
	Command 2:19:26 [buin $UIN][byte $type 0 0 0 0]
}
proc AskAuthorization {UIN Msg} {
	Command 2:19:24 [buin $UIN][pstr [ToWire $Msg [query icq::data -encoding $UIN]]][word 0]
}

array set ListCmd {contacts 2:3:4 visible 2:9:5 invisible 2:9:7}

proc UploadList {type ContactList} {
	set pkg {}
	foreach Contact $ContactList { append pkg [buin $Contact] }
	Command $icq::ListCmd($type) $pkg 
}

proc SetStatus {status {notrec 1}} {
	upvar #0 icq::data data
	variable alias
	log 5 "Setting status $status"
	if {$notrec} {
		if {$status=="offline"} { 
			CloseConnection icq::data
			return
		}
		set data(-status) $status	
		if {$data(MyStatus)=="offline"} { 
			Connect icq::data $data(-server) $data(-port)
			return
		}	
	}
	if {[info exists alias(status:$status)]} {
		set status [expr ($data(Flags)<<16)+$alias(status:$status)]
		Command 2:1:30 [TLV 2 Status $status ErrCode 0\
			LanDetails [LanDetails $data(ip) $data(port)]]
	}
}
# Calculate LanDetails TLV for extended status
proc LanDetails {ip port} {
	set addr 0
	foreach octet [split $ip .] {set addr [expr $addr*256+$octet] }
	foreach {vmaj vmin vrel} [split [package present icq] .] break
	binary format IIcSISSSSIIIS $addr $port 1 8 0 0 0x50 0 3\
		0xffffffbe 0x[format %02x%02x%02x%02x $vmaj $vmin $vrel 0] 0 0
}

proc SendCaps {} {
	Command 2:2:4 [TLV 1 UserInfo [cap ICQ ICQRELAY UTF8 Alicq]]
}


proc SendSMS {Number Message {id 201}} {
	upvar #0 icq::data data
	set Time [clock format [clock seconds] -format "%a, %d %b %Y %T GMT" -gmt yes]
	set message "<icq_sms_message><destination>${Number}</destination><text>${Message}</text>"
	append message "<codepage>1252</codepage><senders_UIN>${data(-uin)}</senders_UIN><senders_name>Bob</senders_name>"
	append message "<delivery_receipt>Yes</delivery_receipt><time>${Time}</time></icq_sms_message>"
	SrvMessage 2000 [bdword 0x82140001 0x00160000 0 0 0 0][lntsz $message] $id
}

proc ChangePassword {pwd {x 2}} {SrvMessage 2000 [lword 1070][lntsz $pwd] $x}
proc ShortInfoRequest {UIN {x 2}} {SrvMessage 2000 [lword 1210][ldword $UIN] $x}
proc FullInfoRequest {UIN {x 2}} {SrvMessage 2000 [lword 1202][ldword $UIN] $x}

proc Search {SearchList {id 204}} {
	array set Attrs $SearchList
	set en $::icq::data(-encoding)
	foreach {formatter item} {lntsz FirstName lntsz LastName
			lntsz Nick lntsz Email lword MinAge lword MaxAge
			byte Sex byte Language lntsz City lntsz State
			lword Country lntsz CompanyName lntsz CompanyDep 
			lntsz CompanyPos byte CompanyOcc bword PastInfoCat
			lntsz Interests lword OrgCat lntsz OrgDesc
			bword HomepageCat lntsz Homepage byte OnlineOnly} {
		if {![info exists Attrs($item)]} {set Attrs($item) {}}
		if {$formatter=="lntsz"} {
			append search_rq [$formatter [ToWire $Attrs($item) $en]]
		} else { 
			if {$Attrs($item)=={}} {set Attrs($item) 0}
			append search_rq [$formatter $Attrs($item)]
		}
	}
	SrvMessage 2000 [lword 1331]$search_rq $id
	unset Attrs search_rq
}

proc UpdateInfo {Info {id 205}} {
	array set Attrs $Info
	foreach {formatter var item} { lntsz main Nick lntsz main FirstName 
		lntsz main LastName lntsz main email lntsz main City 
		lntsz main State lntsz main Phone lntsz main Fax 
		lntsz main Street lntsz main Mobile lntsz main Zip 
		lword main Country byte main TimeZone byte main PublishEmail
		byte more Age byte more Unk byte more Sex lntsz more Homepage
		lword more Year byte more Month byte more Day byte more Lang1
		byte more Lang2 byte more Lang3} {
		if {![info exists Attrs($item)]} {set Attrs($item) {}}
		if {$formatter=={lntsz}} {
			append info($var) [$formatter [ToWire $Attrs($item) $::icq::data(-encoding)]]	
		} else { 
			if {$Attrs($item)=={}} {set Attrs($item) 0}
			append info($var) [$formatter $Attrs($item)]
		}
	}
	foreach {item cmd} {main 1002 more 1021} {
		SrvMessage 2000 [lword $cmd]$info($item) $id
	}	
	unset Attrs
}

proc SrvMessage {cmd packet {id 2}} { 
	upvar #0 icq::data data
	Command 2:21:2 [TLV srv 1 [lnts [ldword $data(-uin)][lword $cmd][lword $id]$packet]] $id 
}

# End of client commands

# Helper functions

# Handle keepalive requests
proc KeepAliveStart {interval} {
	Packet 5
	Ping
	after $interval [nc KeepAliveStart $interval]
}

proc KeepAliveStop {int} { after cancel [nc KeepAliveStart $int] }

# Ping periodically sends SetStatus command to detect if network alive
proc Ping {} {
	upvar #0 icq::data data
	# Send small setstatus packet as ping
	Command 2:1:30 [TLV 2 Status\
		[expr ($data(Flags)<<16)+$icq::alias(status:$data(-status))]]
}

# Force disconnect on ping timeout
proc PingTimeout {} {
	upvar #0 icq::data data
	CloseConnection icq::data
	if {[string is true $data(-reconnect)]} {
		after 10000 [nc Connect icq::data $data(-server) $data(-port)]
	}
}

proc RestartTimer {} {
	upvar #0 icq::data data
	after cancel icq::PingTimeout
	after [expr $data(keepalive)*2] icq::PingTimeout
}

proc FromWire {text enc} {
	regsub -all {\r\n} [encoding convertfrom $enc $text] \n text
	set text
}
proc ToWire {text enc} {
	regsub -all {[\r\n]} [encoding convertto $enc $text] \r\n text
	set text
}

proc query {name k {uin {}}} {
	upvar #0 $name data
	if {[info exists data($uin$k)]} {set data($uin$k)} else {set data($k)}
}

proc Packet {channel {Data {}}} {
	variable SeqNum
	incr SeqNum
	if { $SeqNum>65535 } {set SeqNum 0}
	Send [byte 42 $channel][bword $SeqNum][pstr $Data]
}

proc Connect {name Server Port} {
	upvar #0 $name data
	if {!$data(-uin) || $data(-password)=={}} {
		$data(-event) Error 0:20 "UIN or password is not set."
		return
	}
	CloseConnection $name
	if {$data(-proxy)!={}} { $data(-proxy) connect $Server $Port; return }
	log 1 "Connecting to $Server:$Port"
	if {[catch {set data(socket) [socket -async $Server $Port]}]} {
		$data(-event) Error 0:1 "Unknown host"
		return
	}
	fconfigure $data(socket) -blocking no -translation binary
	log 1 "Name resolved, connecting..."
	fileevent $data(socket) readable [nc ReadData $name]
}

proc ProxyResult {name code sock} {
	upvar #0 $name data
	if {!$code} {
		set data(socket) $sock
		fileevent $data(socket) readable [nc ReadData $name]
	} else {
		CloseConnection $name
		log 0 "Proxy error: code $code"
		if {$code!=403 && [string is true $data(-reconnect)]} {
			after 5000 [nc Connect $name $data(-server) $data(-port)]
		} else {
			$data(-event) Error 0:10 "Proxy error: code $code"
		}
	}
}

proc CloseConnection {name} {
	upvar #0 $name data
	if {[string is true $data(-keepalive)]} {
		KeepAliveStop $data(keepalive)
	}	
	if {[info exists data(socket)]} {
		close $data(socket)
		unset data(socket)
	}
	set data(wantBody) 0
	set data(Length) 6
	set data(MyStatus) offline
	$data(-event) MyStatus offline
}

proc Send {packet} {
	upvar #0 icq::data data
	set full [llength $data(send_queue)]
	lappend data(send_queue) $packet
	binary scan $packet H* hex; log 10 "-> $hex"
	if {!$full} Sender
}

proc Sender {} {
	upvar #0 icq::data data
	if {[catch { puts -nonewline $data(socket) [lindex $data(send_queue) 0]
		     flush $data(socket) }]} {
		$data(-event) Error 0:1
	}
	set data(send_queue) [lrange $data(send_queue) 1 end]
	if {$data(send_queue)!={}} { after idle [nc Sender] }
}

proc EncryptPassword {pwd} {
	set out {}
	foreach in [split $pwd {}] val {243 38 129 196 57 
		134 219 146 113 163 185 230 83 122 149 124} {
		if {$in=={}} break
		binary scan $in c char
		append out [binary format c [expr $char^$val]]
	}
	set out
}

# Make proper SNAC header
proc SNAC {Family Sub FlagA FlagB ReqId} {
	binary format SSccSS $Family $Sub $FlagA $FlagB $ReqId $Sub
}

proc Command {Cmd Data {ReqId 0}} {
	foreach {ChanId Family SubFamily} [split $Cmd :] break
	Packet $ChanId [SNAC $Family $SubFamily 0 0 $ReqId]${Data}
}

proc nc {args} { namespace code $args }
proc alias {key val} {
	variable alias
	if {[info exists alias($key:$val)]} {set val $alias($key:$val)}
	set val
}
# UCS-2BE support appeared in Tcl 8.4 only, so handle this encoding manually
# by converting into UCS-2LE
proc swab {str} {
	foreach {x y} [split $str {}] {append res $y$x}
	set res
}
# ICQ Types 
# New typingG procs
# Generic string with length at the beginning template 
proc genstr.s {type val} {binary format ${type}a* [string length $val] $val}
proc genstr.g {type str} {
	binary scan $str $type len
	binary scan $str ${type}a${len}a* _ res tail
	list $res $tail
}
# string with zero at the end, having length in the beginning (word)
proc lntsz.set {val} { lnts.set [format %s%c $val 0]}
proc lntsz.get {str} {
	set r [lnts.get $str]
	list [string range [lindex $r 0] 0 end-1] [lindex $r 1]
}
# Intergal types (those supported by ``binary'' command) template
proc getintegral {spec str} {
	binary scan $str ${spec}a* res other
	list $res $other
}
proc integral {spec args} { binary format ${spec}[llength $args] $args }
# Capability: 16-byte hex value
proc cap.get {str} { 
	binary scan $str H32a* res str
	list [alias cap $res] $str
}
proc cap.set {val} { binary format H32 [alias cap $val]}
# Capability in the 'cacnonical' form
proc guid.get {str} { alias cap [string tolower [string map {\{ {} \} {} - {}} $str]] }
proc guid.set {val} {
	set val [alias cap $val]
	string toupper "{[string range $val 0 7]-[string range $val 8 11]-[string range $val 12 15]-[string range $val 16 19]-[string range $val 20 end]}"
}

# Get TLVs of channel 2
proc TLV.get {str} {
	binary scan $str Sa* count str
	scanTLV $str $count 2
}

# Build binary packed list of TLVs
proc TLV {channel args} {
	variable tlvSpec
	set res {}
	foreach {tag value} $args {
		set tag [alias $channel $tag]
		if {[info exists tlvSpec($channel:$tag)]} {
			 set spec $tlvSpec($channel:$tag)
		} else { set spec a* }
		set p [binary format $spec $value]
		append res [binary format SS $tag [string length $p]]$p
	}
	set res
}

# Get list of TLVs as pair-volume list
proc scanTLV {str count channel} {
	variable tlvSpec
	set res [list]
	while {$count && $str!={}} {
		binary scan $str SSa* tag len str
		binary scan $str a${len}a* value str
		if {[info exists tlvSpec($channel:$tag)]} {
			binary scan $value $tlvSpec($channel:$tag) value
		}
		lappend res [alias $channel $tag] $value
		incr count -1
	}
	list $res $str
}

proc *.get {val} { list $val {} }

proc match {str spec} {
	foreach item $spec {
		set type [lindex $item 0]
		if {[string is integer $type]} {
			binary scan $str @${type}a* str
			continue
		}
		if {![llength [info commands ${type}.get]]} {
			return -code error "Unknown type $type"
		}
		set vars [lrange $item 1 end]
		if {![llength $vars]} { set vars [list {}] }
		foreach v $vars {
			upvar $v res
			foreach {res str} [${type}.get $str] break
		}
	}
}
# Make aliases for type creation procs
proc inject {cmd res args} {
	foreach x $args { append res [$cmd $x] }
	set res
}

set ns [namespace current]

# Spawn set of aliases to represent differnet kinds of strings
foreach {name type} {buin c lnts s pstr S dstr i} {
	interp alias {} ${ns}::${name}.set {} ${ns}::genstr.s $type
	interp alias {} ${ns}::${name}.get {} ${ns}::genstr.g $type
}
# Spawn set of aliases to simplify setting several values of same type
foreach item [info commands *.set] {
	interp alias {} ${ns}::[string range $item 0 end-4] {} ${ns}::inject $item {}
}
# Spawn set of aliases to represent intergal types: word, dword, byte etc
foreach {key val} {byte c word s lword s bword S dword i ldword i bdword I} {
	interp alias {} ${ns}::$key {} ${ns}::integral $val
	interp alias {} ${ns}::$key.get {} ${ns}::getintegral $val
}
#
# Create template of login TLV
foreach {key val} {
	ClientProfile {ICQ Inc. - Product of ICQ (TM).2000b.4.63.1.3279.85}
	ClientType 266 VersionMaj 4 VersionMin 63 IcqNumber 1
	BuildMaj 3279 BuildMin 85 Language en Country us} {
		append loginTLV [TLV 1 $key $val]
}
}

