package provide proxy 0.1

proc nonempty {var} {
   upvar 1 $var v	
   expr {[info exists v] && ![string match $v {}]}
}

namespace eval proxy {

set counter 0
proc _logger {level msg} {}
proc types {} { map x [namespace children] {namespace tail $x} }

proc proxy {args} {
	set id data${::proxy::counter}
	incr proxy::counter
	upvar #0 ::proxy::$id data
	set data(log) proxy::_logger
	foreach {key val} $args { configure $id $key $val }
	foreach key {type server port} {
		if {![info exists data($key)]} {
			return -code error "$key is not specified for proxy"
		}
	}
	interp alias {} $id {} proxy::proxycmd $id
	set id
}

proc configure {id key args} {
	upvar #0 ::proxy::$id data
	if {![string match -* $key]} { return -code error "Invalid key $key" }
	set k [string range $key 1 end]
	if {[lsearch {type server port user password result log} $k]==-1} {
		return -code error "Invalid key $key"
	}
	if {![llength $args]} {return $data($k)}
	set val [lindex $args 0]
	if {[string equal $k type] &&\
	    ![llength [namespace children [namespace current] $val]]} {
		return -code error "Proxy type $val is not supported.\nPossible values are: [join [types] {, }]"
	}
	set data($k) $val
}

proc proxycmd {id cmd args} {
	upvar #0 proxy::$id data 
	if {$cmd=="configure"} {return [eval proxy::configure $id $args]}
	if {[llength [info procs ::proxy::$data(type)::$cmd]]!=1} {
		return -code error "$cmd is not supported.\nPossible values are: [info procs ::proxy::$data(type)::*]"
	}
	eval ::proxy::$data(type)::$cmd $id $args
}

# Http proxy
namespace eval http {
proc connect {id Server Port} {
	upvar #0 proxy::$id data 
	if {[nonempty data(user)]} { uplevel #0 "package require base64" }
	$data(log) 3 "Connecting to proxy $data(server):$data(port)"
	if {[catch {set data(socket) [socket -async $data(server) $data(port)]} v]} { return -code error "Can't connect to proxy: $v" }
	fconfigure $data(socket) -translation crlf -buffering line -blocking no
	fileevent $data(socket) writable "proxy::http::write $id $Server $Port"
}

proc write {id Server Port} {
	upvar #0 proxy::$id data 
	$data(log) 5 "CONNECT $Server HTTP/1.0"
	set msg "CONNECT $Server HTTP/1.0"
	if {[nonempty data(user)]&&[nonempty data(password)]} {
		set auth [::base64::encode $data(user):$data(password)]
		append msg "\n" "Proxy-Authorization: Basic $auth"
	}
	if {[catch {
		puts $data(socket) "$msg\n"
		flush $data(socket)
		fileevent $data(socket) writable {}
		fileevent $data(socket) readable "proxy::http::read $id"
	   }]} {$data(result) -1 0 }	
}

proc read {id} {
	upvar #0 proxy::$id data
	set count [gets $data(socket) http_resp]
	if {$count==-1} {
		if {$http_resp!={}} return 
		fileevent $data(socket) readable {}
	}
	if {$count==0} {
		if {[regexp {^HTTP\/1\.[01] ([0-9]+)} $data(Body) s http_status]} {
			if {$http_status==200} {
				fconfigure $data(socket) -blocking no -translation binary
				$data(result) 0 $data(socket)
				$data(log) 3 "Connected to proxy"
			} else {
				fileevent $data(socket) readable {}
				$data(result) $http_status 0]
			}
		}
	} else { append data(Body) $http_resp "\n"; return }
}
}
}
